/*****************************************************************************
 * $LastChangedDate: 2009-11-22 21:39:11 -0600 (Sun, 22 Nov 2009) $
 * @file
 * @author  Jim E. Brooks  http://www.palomino3d.org
 * @brief   Helper functions for STL.
 *//*
 * LEGAL:   COPYRIGHT (C) 2004 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#ifndef BASE_FUNCS_STL_HH
#define BASE_FUNCS_STL_HH 1

#include <algorithm>

namespace base {

/*****************************************************************************
 * Get raw C array from STL vector.
 * &v[0] is a STL idiom.
 *****************************************************************************/
template<typename T>       T*       PTR(       vector<T>& v ) { return &v[0]; }
template<typename T> const T* CONST_PTR( const vector<T>& v ) { return &v[0]; } 

/*****************************************************************************
 * Expand a sequential STL container if necessary.
 * Useful before operator[] to ensure container can hold it.
 *****************************************************************************/
template<typename SEQCONT>
void Expand( SEQCONT& seqcont, uint idx )
{
ASSERT2( idx < MAX_ELEMS );

    // Be careful, resize() can implode a container too.
    if ( UX( idx >= seqcont.size() ) )
        seqcont.resize( idx + 1 );  // one more than index
}

/*****************************************************************************
 * Remove() is a compatible way to remove items from various STL containers.
 * Does NOT destroy/delete items.
 *****************************************************************************/
template<typename T>
void
Remove( vector<T>& container, T item )  // STL vector
{
    // erase/remove idiom for STL vectors or deques.  erase() returns nothing.
    container.erase( std::remove( container.begin(), container.begin(), item ),
                     container.end() );
}

template<typename T>
void
Remove( deque<T>& container, T item )  // STL deque
{
    container.erase( std::remove( container.begin(), container.begin(), item ),
                     container.end() );
}

template<typename T>
void
Remove( list<T>& container, T item )  // STL list
{
    container.remove( item );
}

template<typename KEY>
void
Remove( set<KEY>& container, KEY item )  // STL set
{
    container.erase( item );
}

template<typename KEY,typename VAL>
void
Remove( map<KEY,VAL>& container, VAL item )  // STL map
{
    container.erase( item );
}

/*****************************************************************************
 * @return True if two STL containers have duplicates.
 *****************************************************************************/
template<typename CONTAINER>
bool
IfDuplicate( const CONTAINER& c1, const CONTAINER& c2 )
{
    for ( typename CONTAINER::const_iterator i1 = c1.begin(); i1 != c1.end(); ++i1 )
    {
        if ( std::find( c2.begin(), c2.end(), *i1 ) != c2.end() )
            return true;
    }
    return false;
}

} // namespace base

#endif // BASE_FUNCS_STL_HH
