/*****************************************************************************
 * $LastChangedDate: 2009-11-22 21:39:11 -0600 (Sun, 22 Nov 2009) $
 * @file
 * @author  Jim E. Brooks  http://www.palomino3d.org
 * @brief   Macros (excluding function macros).
 *//*
 * LEGAL:   COPYRIGHT (C) 2004 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#ifndef DOXYGEN  // doxygen can't parse this correctly
#ifndef BASE_MACROS_HH
#define BASE_MACROS_HH 1

namespace base {

////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////  macros  /////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

// Prevent copying an object.  Put these in the private section of a class.
// The copy methods are prototypes without definitions, so usage causes a linker error.
#define PREVENT_COPYING( CLASS )                    \
private:                                            \
    CLASS( const CLASS& );                          \
    CLASS& operator=( const CLASS& );

// To mark classes that are copyable.
#define COPYABLE( CLASS )          // means class is copyable by user-defined copying methods
#define COPYABLE_DEFAULT( CLASS )  // means class is copyable by compiler-generated copying methods
#define ALLOW_COPYING( CLASS )     // synonym

// To mark class that shouldn't be instantiated
#define PREVENT_INSTANTIATING( CLASS )  // NOP
#define PREVENT_INSTANCE( CLASS )  // NOP

/// Helper for correctly writing operator<() for STL container.
/// bool operator<( const C& a, const C& b )
/// {
///     RETURN_LT_TRUE_GT_FALSE( a.x, b.x )
///     RETURN_LT_TRUE_GT_FALSE( a.y, b.y )
///     return false;  // equal
/// }
/// "Effective C++" Item 21 says that a comparison function
/// should return false for equal values or it will corrupt a set/map.
#define RETURN_LT_TRUE_GT_FALSE( X, Y )         \
{{                                              \
    if ( (X) < (Y) ) return true;               \
    if ( (Y) < (X) ) return false;              \
}}

// Put this in destructors of globals which may call other globals
// to avoid pitfall of undefined order of destruction.
#define RETURN_IF_EXITING  {{ if ( base::gGlobals.mExiting ) return; }}

// Arcane indirect trick to give a variable a unique name.
// Subtlety: If written on the same line, UNIQUE_NAME() expands the same name.
// That's a feature not a bug.  BEGIN_ONCE depends on it.
#define UNIQUE_NAME__(NAME,LINE) NAME##LINE
#define UNIQUE_NAME_(NAME,LINE)  UNIQUE_NAME__(NAME,LINE)
#define UNIQUE_NAME(NAME)        UNIQUE_NAME_(NAME,__LINE__)

/// @def BEGIN_ONCE
// BEGIN_ONCE {  <-- write brace
// ...
// } END_ONCE
// Subtlety: This must be one-line because of UNIQUE_NAME().
#define BEGIN_ONCE                          \
{                                           \
    PERSISTENT int UNIQUE_NAME(once);       \
    ++UNIQUE_NAME(once);                    \
    if ( UNIQUE_NAME(once) == 1 )

#define END_ONCE \
}

/// @def likely
/// Branch hint for compiler.
#if __GNUC__ >= 3
#define likely(COND)    __builtin_expect((COND),1)
#define unlikely(COND)  __builtin_expect((COND),0)
#else
#define likely(COND) (COND)
#define unlikely(COND) (COND)
#endif
#define EX(COND) likely((COND))   ///< expected condition
#define UX(COND) unlikely((COND)) ///< unexpected condition

/// @def UNUSED
/// Tell compiler a var is unused.
#if GCC_VERSION >= 30100  // if gcc-3.1 or later
#define UNUSED    __attribute__ ((unused))
#define DEBUG_VAR __attribute__ ((unused))
#define NOINLINE  __attribute__ ((noinline))
#else
#define UNUSED
#define DEBUG_VAR
#define NOINLINE
#endif

/// Expand inline singleton method.
#define DEF_GetInstance( CLASS ) \
    CLASS_METHOD CLASS& GetInstance( void ) \
    { \
        if ( UX( msInstance == NULL ) ) \
            msInstance = new CLASS; \
        return *msInstance; \
    }

} // namespace base

#endif // BASE_MACROS_HH
#endif // DOXYGEN
