/******************************************************************************
 * $LastChangedDate: 2010-01-30 15:48:12 -0600 (Sat, 30 Jan 2010) $
 * @file
 * @author  Jim E. Brooks  http://www.palomino3d.org
 * @brief   C++ streams.
 *//*
 * LEGAL:   COPYRIGHT (C) 2004 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 ******************************************************************************/

#define BASE_STREAM_CC 1
#include "base/module.hh"
#include "base/stream.hh"
#include "base/file.hh"
#undef cout
#undef cerr

namespace base {

////////////////////////////////////////////////////////////////////////////////
/////////////////////////////  functions  //////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/*****************************************************************************
 * Open a stream to a file.
 * @param   fs
 *          File stream.
 * @param   bits
 *          Open for reading and/or writing.
 * @param   filename
 * @return  False if stream error.
 *****************************************************************************/
bool
StreamOpen( fstream& fs, ios::openmode bits, const string& filename )
{
    // Binary else trouble.
    bits |= ios::binary;

    // Try opening a stream to a file.
    // For the open-for-writing case, first try to open the file without the truncate bit.
    // But some iostream implementations might fail if the file doesn't exist.
    fs.open( filename.c_str(), bits );
    if ( (not fs.good())                // if error
     and (bits & ios::out)              // if output stream
     and (not (bits & ios::trunc)) )    // if truncate not already set
    {
        // This might be the file-doesn't-exist case depending
        // on the iostream implementation.  Retry with truncate bit.
        fs.clear();
        fs.open( filename.c_str(), bits | ios::trunc );
    }

    return fs.good();
}

/*****************************************************************************
 * Read all chars to EOL.
 * @return False if stream error.
 *****************************************************************************/
bool
StreamReadLine( istream& strm, string& line )
{
    line.erase();
    char c;
    while ( strm.good() and (c = strm.get()) != '\n' )
    {
        line += c;
    }
    return strm.good();
}

/*****************************************************************************
 * Jump to the next line.
 * @return False if stream error.
 *****************************************************************************/
bool
StreamToNextLine( istream& strm )
{
    while ( strm.good() && (strm.get() != '\n') ) { }
    return strm.good();
}

/*****************************************************************************
 * Peek input stream.
 * @return True if word is in input stream.
 *****************************************************************************/
bool
StreamPeek( istream& strm, const string& word )
{
    bool result = false;

    // Save read position.
    std::ios::pos_type pos = strm.tellg();

    try
    {
        // Read char-by-char into a buf
        string buf;
        for ( uint i = 0; i < word.size(); ++i )
        {
            const char c = char(strm.get());
            if ( not strm.good() )
                break;

            switch ( c )
            {
                case CHAR_LF: case CHAR_CR: case ' ': case '\t':
                {
                    --i;  // don't append whitespace to buf[]
                    continue;
                }
                break;

                default:
                {
                    buf += char(c);
                }
                break;
            }

            if ( buf[i] != word[i] )  // mismatch?
                break;
        }
        result = (buf == word) && strm.good();
    }
    catch (...)
    {
        result = false;
    }

    // Restore read position.
    strm.seekg( pos );

    return result;
}

} // namespace base
