/*****************************************************************************
 * $LastChangedDate: 2009-11-22 17:05:48 -0600 (Sun, 22 Nov 2009) $
 * @file
 * @author  Jim E. Brooks  http://www.palomino3d.org
 * @brief   Timer class that invokes registered callbacks.
 *//*
 * LEGAL:   COPYRIGHT (C) 2006 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#ifndef BASE_TIMER_HH
#define BASE_TIMER_HH 1

#define GET_TIMER() (base::Timer::GetInstance())

#include "base/time.hh"  // Milliseconds etc

namespace base {

const bool TIMER_ENABLED = true;
const bool TIMER_DISABLED = false;

// Macro to define nested class.
// Functor must not hold a ref-ptr to its owner
// (would create a cycle of reference-counted pointers).
#define CLASS_TICK_FUNCTOR( CLASS, MEMBER )                             \
    class TickFunctor : public base::Timer::Functor                     \
    {                                                                   \
    public:                                                             \
        TickFunctor( CLASS& obj ) : MEMBER(obj) { }                     \
        void operator()( const Milliseconds millisecElapsed );          \
    private:                                                            \
        CLASS& MEMBER;                                                  \
    };                                                                  \
    friend class TickFunctor;

// Macro.
#define CLASS_TICK_FUNCTOR_BASIC                                        \
    class TickFunctor : public base::Timer::Functor                     \
    {                                                                   \
    public:                                                             \
        void operator()( const Milliseconds millisecElapsed );          \
    };                                                                  \
    friend class TickFunctor;

///////////////////////////////////////////////////////////////////////////////
/// @brief Singleton to register, unregister, and invoke timer-tick callbacks.
///
/// Trying to re-register the same functor is considered a bug.
///
/// A callback is allowed to unregister itself,
/// even while Timer is iterating thru callbacks.
/// Unregister() will mark-for-destruction a busy callback,
/// (or will immediately destroy it if no iteration is happening).
///
/// Disabling Timer isn't absolute.
/// A callback that is constructed with Timer::CANNOT_DISABLE
/// must still be called even if Timer is disabled.
/// For example, polling keyboard should remain enabled always.
///
/// To integrate Timer:
/// Whichever function that is called periodically must call Timer::Tick()
/// to drive the Timer object.  Otherwise Timer won't be active.
///
/// Timer must be Singleton since globals register timers.
///
/// Pausing program by disabling Timer:
/// -----------------------------------
/// Time is stopped while the program is paused, except for
/// callbacks that cannot be disabled (still called while paused).
///
/// GetElapsedTime() returns the elapsed time minus time lost while paused.
/// GetElapsedTimeTotal() still counts time while Timer is disabled (paused).
/// Callbacks that cannot be disabled should call GetElapsedTimeTotal()
/// else their timing calculations will malfunction after pausing.
///
class Timer
{
PREVENT_COPYING(Timer)
private:
    enum eRegistered { CALLBACK_UNREGISTERED = 0,  // "REGISTERED" predefined on Windows
                       CALLBACK_REGISTERED   = 1 };

public:
    enum eDisablement { CAN_DISABLE, CANNOT_DISABLE };

    ////////////////////////////////////////////////////////////////////////////////
    /// @brief User-defined timer functor.
    ///
    /// Functors are passed as a shptr, not by value, in case a client modifies
    /// the Functor without having a copy ctor (avoids modifying-a-copy pitfall).
    ///
    class Functor : public Shared
    {
    PREVENT_COPYING(Functor)
    public:
                        Functor( void ) { }
        virtual         ~Functor() { }
        /// millsecElapsed is time when program started and now.
        virtual void    operator()( const Milliseconds millisecElapsed ) = 0;
    };

private:
    ////////////////////////////////////////////////////////////////////////////
    /// @brief Encapsulates a callback functor and defines how frequently to call it.
    ///
    /// Callbacks are modified with the last time they were called.
    /// Callback has a copy ctor to avoid the modifying-a-copy pitfall
    /// if STL copies them (relic from using a list).
    ///
    class Callback : public Shared
    {
    COPYABLE(Callback)
    friend class Timer;
    public:
                    Callback( const bool registered,
                              const shptr<Functor> functor,
                              const Milliseconds millisecFreq,
                              const Timer::eDisablement canDisable = CAN_DISABLE );
                    Callback( const Callback& src );
                    ~Callback();
        void        CallIfTimeHasCome( const Milliseconds millisecElapsed );
        bool        IfRegistered( void ) const { return mRegistered; }
        bool        IfCanDisable( void ) const { return mCanDisable; }

    private:
        enum                { NEVER_CALLED = -1 };
        bool                mRegistered;        ///< false if Unregister() applied
        shptr<Functor>      mFunctor;           ///< the callback
        Milliseconds        mMillisecFreq;      ///< how often to invoke callback
        Milliseconds        mMillisecCalled;    ///< last time it was called
        bool                mCanDisable;        ///< false if callback must be called even if Timer is disabled
    public:
        DECL_TYPESIG(TYPESIG_TIMER_CALLBACK)
    };

private:
    Timer( void );  // singleton
    ~Timer();

public:
// Client interface:
                        DEF_GetInstance( Timer )  // Singleton
    void                Enable( bool enable );

    // Time since start of program (minus time lost while program was paused).
    Milliseconds        GetElapsedTime( void ) const { return mVirtualTime; }

    // Total time since start of program.
    Milliseconds        GetElapsedTimeTotal( void ) const { return mTotalTime; }

    void                Register( shptr<Functor> functor, const Milliseconds millisecFreq, eDisablement canDisable = CAN_DISABLE );
    void                Unregister( shptr<Functor> functor );

// For integration:
    void                Tick( void );  // to pulse Timer

private:
    typedef map<SafePtr<Functor>,Callback> Callbacks;
    Callbacks::iterator Find( SafePtr<Functor> functor );

private:
    CLASS_CONST uint    MIN_MILLISEC = 1;
    CLASS_CONST uint    MAX_MILLISEC = THOUSAND * 60 * 60;  // 1 hour
    CLASS_VAR Timer*    msInstance;
    bool                mEnabled;                       ///< program will be (mostly) paused if Timer is disabled
    bool                mBusy;                          ///< if iterating thru callbacks
    Milliseconds        mTotalTime;                     ///< total time since start of program
    Milliseconds        mVirtualTime;                   ///< time since start of program minus time lost while paused
    Milliseconds        mTotalTimeWhenPauseStarted;     ///< copy of mTotalTime when pause was started
    Milliseconds        mVirtualTimeWhenPauseStarted;   ///< copy of mVirtualTime when pause was started
    Milliseconds        mLostTimeWhilePaused;           ///< accumulated time lost while paused
    Callbacks           mCallbacks;                     ///< mapped by raw ptr to Functor
};

} // namespace base

#endif // BASE_TIMER_HH
