/*****************************************************************************
 * $LastChangedDate: 2009-11-22 21:39:11 -0600 (Sun, 22 Nov 2009) $
 * @file
 * @author  Jim E. Brooks  http://www.palomino3d.org
 * @brief   Collision-detection class.
 *//*
 * LEGAL:   COPYRIGHT (C) 2008 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#define COLLISION_COLLISION_CC 1
#include "base/module.hh"
#include "base/conf.hh"
using namespace base;
#include "math/module.hh"
#include "math/funcs.hh"
#include "math/funcs_vector.hh"
using namespace math;
#include "object/module.hh"
#include "object/dyna.hh"
#include "object/events.hh"
using namespace object;
#include "world/module.hh"
using namespace world;
#include "collision/module.hh"
#include "collision/collision.hh"
#include "collision/defs.hh"
#include "collision/events.hh"

#include <osgUtil/LineSegmentIntersector>
#include <osgUtil/IntersectionVisitor>

namespace collision {

////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////  Collision  //////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

Collision* Collision::msInstance;

/*****************************************************************************
 * ctor/dtor.
 *****************************************************************************/
Collision::Collision( void )
:   mCollidables()
{
    // Unlisten() isn't necessary for singletons.
    EVENT_DESTROY_OBJECT.Listen( new DestroyObjectListener( *this ) );
}

Collision::~Collision()
{
    // NOP
}

/*****************************************************************************
 * Set an Object as collidable or indestructible.
 *****************************************************************************/
void
Collision::SetCollidable( shptr<Object> object, const bool collidable )
{
CHECK_TYPESIG(object,TYPESIG_OBJECT);

    if ( collidable )
        mCollidables.insert( object );
    else
        mCollidables.erase( object );
}

/*****************************************************************************
 * @return (true,object) or (false,NULL).
 *
 * Compute intersections of 2 rays:
 * - ray of "down vector" for contact force
 * - ray of the movement of Dyna
 *****************************************************************************/
pair<bool,shptr0<Object> >
Collision::IfCollision( shptr<Dyna> dyna )
{
CHECK_TYPESIG(dyna,TYPESIG_DYNA);

    // Is this Dyna indestructible?
    // A special-case is to ignore zombie Dynas with a minimal graph.
    // SUBTLE: IfCollidable() will return false for "zombie" Objects
    // (which one-node "minimal graph" that will cause OSG intersection-test to fail).
    if ( not dyna->IfCollidable() )
        return make_pair<bool,shptr0<Object> >( false, NULL );

    // Group of 1 or 2 LineIntersectors.
    RefPtr<osgUtil::IntersectorGroup> intersectorGroup = new osgUtil::IntersectorGroup;

    // Compute ray of "down vector" for contact force.
    const WorldVertex worldUpVector  = GET_WORLD().ComputeUpVector( *dyna );  // normalized
    const fp          dynaHeightHalf = dyna->GetHeight() * 0.5f;
    const WorldVertex dynaCenter = dyna->GetPosition();
    const WorldVertex dynaTop(    dynaCenter[XX] + worldUpVector[XX] * dynaHeightHalf,
                                  dynaCenter[YY] + worldUpVector[YY] * dynaHeightHalf,
                                  dynaCenter[ZZ] + worldUpVector[ZZ] * dynaHeightHalf );
    const WorldVertex dynaBottom( dynaCenter[XX] - worldUpVector[XX] * dynaHeightHalf,
                                  dynaCenter[YY] - worldUpVector[YY] * dynaHeightHalf,
                                  dynaCenter[ZZ] - worldUpVector[ZZ] * dynaHeightHalf );
    intersectorGroup->addIntersector( new osgUtil::LineSegmentIntersector( dynaTop,
                                                                           dynaBottom ) );

    // Compute ray of movement of Dyna (skip if Dyna hasn't moved far enough).
    const pair<WorldVertex,WorldVertex> step = dyna->GetTranslationStep();
    const WorldVertex& translationBegin      = step.first;
    const WorldVertex& translationEnd        = step.second;
    const fp           translationDistance   = Distance( translationBegin, translationEnd );
    if ( translationDistance > world::conv::Meters2Sim( collision::defs::MIN_TRANSLATION_DISTANCE_METERS ) )
    {
        intersectorGroup->addIntersector( new osgUtil::LineSegmentIntersector( translationBegin,
                                                                               translationEnd ) );
    }

    // Attach Intersectors to IntersectionVisitor.
    osgUtil::IntersectionVisitor visitor( intersectorGroup.get() );

    // For each collidable Object.
    for ( Collidables::const_iterator iter = mCollidables.begin();
          iter != mCollidables.end();
          ++iter )
    {
        // Get collidable.
        shptr<Object> collidable = *iter;
        CHECK_TYPESIG(collidable,TYPESIG_OBJECT);

        // Object should not crash into itself!
        // And ignore crashes into zombies (zombies will linger temporarily).
        if ( EX( not collidable->IfZombie() and dyna.PTR() != collidable.PTR() ) )
        {
            // Detect collision using by visiting this Object's nodes.
            collidable->GetRootNode()->accept( visitor );

            // Collision?
            if ( intersectorGroup->containsIntersections() )
            {
                // Broadcast event.
                EVENT_COLLISION.Broadcast( make_pair< shptr<Object>, shptr<Object> >( dyna.PTR(), collidable ) );

                return make_pair<bool,shptr0<Object> >( true, collidable.PTR() );
            }
        }

        // Reset intersection.
        visitor.reset();
        intersectorGroup->reset();
    }

    // No collision.
    return make_pair<bool,shptr0<Object> >( false, NULL );
}

/*****************************************************************************
 * Listen for the DestroyObject event.
 *****************************************************************************/
void
Collision::DestroyObjectListener::operator()( shptr<Object> object )
{
CHECK_TYPESIG(object,TYPESIG_OBJECT);

    mCollision.mCollidables.erase( object );
}

} // namespace collision
