/*****************************************************************************
 * $LastChangedDate: 2009-11-22 21:39:11 -0600 (Sun, 22 Nov 2009) $
 * @file
 * @author  Jim E. Brooks  http://www.palomino3d.org
 * @brief   Commands to control an Aircraft.
 *//*
 * LEGAL:   COPYRIGHT (C) 2008 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#define CONTROL_COMMAND_CC 1
#include "base/module.hh"
using namespace base;
#include "math/module.hh"
#include "math/funcs_debug.hh"
#include "math/quaternion.hh"
using namespace math;
#include "world/module.hh"
using namespace world;
#include "control/module.hh"
#include "control/command.hh"

namespace control {

INTERN const string COMMAND_NAME_POSITION           = "pos";
INTERN const string COMMAND_NAME_QUATERNION         = "quat";
INTERN const string COMMAND_NAME_PHYSICAL_ROTATE    = "phyRot";
INTERN const string COMMAND_NAME_THROTTLE           = "throt";
INTERN const string COMMAND_NAME_SPEED              = "kph";
INTERN const string COMMAND_NAME_LANDING_GEAR       = "gear";
INTERN const string COMMAND_NAME_BRAKES             = "brake";

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////  functions  ///////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/*****************************************************************************
 * Write Command (or derivative) to stream.
 * Uses the virtual Command::Write(stream).
 *
 * Asymmetrically, operator<<() (write) is defined, but operator>>() (read) is N/A,
 * since reading a derivative from stream into base is illogical.
 * Rather, use the factory method Command::Read(istream&).
 *****************************************************************************/
ostream& operator<<( ostream& strm, const Command& command )
{
CHECK_TYPESIG(&command,TYPESIG_COMMAND);

    command.Write( strm );  // virtual method
    return strm;
}

////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////  Command  ///////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/*****************************************************************************
 * Factory method.
 * Construct a Command derivative from input stream.
 *****************************************************************************/
shptr<Command>
Command::Read( istream& strm )
{
//CHECK_TYPESIG(this,TYPESIG_COMMAND);  // class method

    // Read name of command.
    // If stream is bad then it'll be empty and won't match anything.

    string cmdName;
    strm >> cmdName;

    if ( cmdName == COMMAND_NAME_POSITION )
    {
        fp x, y, z;
        strm >> x >> y >> z;
        return new CommandPosition( GeoVertex( Meter(x), Meter(y), Meter(z) ) );
    }
    else if ( cmdName == COMMAND_NAME_QUATERNION )
    {
        Quaternion quat;
        strm >> quat;
        return new CommandQuaternion( quat );
    }
    else if ( cmdName == COMMAND_NAME_PHYSICAL_ROTATE )
    {
        uint axis = 0;                      strm >> axis;
        fp controlFraction = 0;             strm >> controlFraction;
        Milliseconds::Int controlFreq = 0;  strm >> controlFreq;
        ASSERT_AXIS(axis);
        return new CommandPhysicalRotate( axis, controlFraction, Milliseconds(controlFreq) );
    }
    else if ( cmdName == COMMAND_NAME_THROTTLE )
    {
        fp arg = 0; strm >> arg;
        return new CommandThrottle( arg );
    }
    else if ( cmdName == COMMAND_NAME_SPEED )
    {
        fp arg = 0; strm >> arg;
        return new CommandSpeed( physics::conv::KPH2Speed( SpeedKPH(arg) ) );
    }
    else if ( cmdName == COMMAND_NAME_LANDING_GEAR )
    {
        bool arg = 0; strm >> arg;
        return new CommandLandingGear( arg );
    }
    else if ( cmdName == COMMAND_NAME_BRAKES )
    {
        bool arg = 0; strm >> arg;
        return new CommandBrakes( arg );
    }
    else
    {
        throw ExceptionCommand( string("Command::Read: stream parse error ") + cmdName );
        return 0;  // shutup compiler
    }
}

////////////////////////////////////////////////////////////////////////////////
//////////////////////////////  CommandPosition  ///////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/*****************************************************************************
 * Set position.
 *****************************************************************************/
CommandPosition::CommandPosition( const GeoVertex& geoPos )
:   mGeoPos(geoPos)
{
}

void
CommandPosition::Execute( shptr<Aircraft> aircraft, const ExecuteMode mode )
{
CHECK_TYPESIG(this,TYPESIG_COMMAND);

    if ( mode & EXECUTE_MODE_GRAPHICS )
    {
        aircraft->SetPosition( world::conv::GeoVertex2WorldVertex( mGeoPos ) );
    }
}

void
CommandPosition::Write( ostream& strm ) const
{
CHECK_TYPESIG(this,TYPESIG_COMMAND);

    // Write position as Meters (GeoVertex).
    strm << COMMAND_NAME_POSITION << ' '
         << mGeoPos[XX].FP() << ' '
         << mGeoPos[YY].FP() << ' '
         << mGeoPos[ZZ].FP() << ' ';  // omit endl
}

// specific:
WorldVertex
CommandPosition::GetWorldVertex( void ) const
{
CHECK_TYPESIG(this,TYPESIG_COMMAND);

    return world::conv::GeoVertex2WorldVertex( mGeoPos );
}

////////////////////////////////////////////////////////////////////////////////
/////////////////////////////  CommandQuaternion  //////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/*****************************************************************************
 * Apply graphical rotation by quaternion.
 * Quaternions only define rotation.
 *****************************************************************************/
CommandQuaternion::CommandQuaternion( const Quaternion& quat )
:   mQuat(quat)
{
}

void
CommandQuaternion::Execute( shptr<Aircraft> aircraft, const ExecuteMode mode )
{
CHECK_TYPESIG(this,TYPESIG_COMMAND);

    if ( mode & EXECUTE_MODE_GRAPHICS )
    {
        aircraft->SetQuaternion( mQuat );
    }
}

void
CommandQuaternion::Write( ostream& strm ) const
{
CHECK_TYPESIG(this,TYPESIG_COMMAND);

    strm << COMMAND_NAME_QUATERNION << ' '
         << mQuat << ' ';  // omit endl
}

////////////////////////////////////////////////////////////////////////////////
///////////////////////////  CommandPhysicalRotate  ////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/*****************************************************************************
 * Rotate according to physics model.
 *****************************************************************************/
CommandPhysicalRotate::CommandPhysicalRotate( uint axis, fp controlFraction, Milliseconds controlFreq )
:   mAxis(axis),
    mControlFraction(controlFraction),
    mControlFreq(controlFreq)
{
ASSERT_AXIS(axis);
}

void
CommandPhysicalRotate::Execute( shptr<Aircraft> aircraft, const ExecuteMode mode )
{
CHECK_TYPESIG(this,TYPESIG_COMMAND);

    if ( mode & EXECUTE_MODE_PHYSICS )
        aircraft->PhysicalRotate( mAxis, mControlFraction, mControlFreq );
}

void
CommandPhysicalRotate::Write( ostream& strm ) const
{
CHECK_TYPESIG(this,TYPESIG_COMMAND);
ASSERT_AXIS(mAxis);

    strm << COMMAND_NAME_PHYSICAL_ROTATE << ' '
         << mAxis << ' '
         << mControlFraction << ' '
         << mControlFreq << ' ';  // omit endl
}

////////////////////////////////////////////////////////////////////////////////
///////////////////////////////  CommandThrottle  //////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/*****************************************************************************
 * Set throttle command.
 *****************************************************************************/
CommandThrottle::CommandThrottle( const fp throttle )
:   mThrottle(throttle)
{
}

void
CommandThrottle::Execute( shptr<Aircraft> aircraft, const ExecuteMode mode )
{
CHECK_TYPESIG(this,TYPESIG_COMMAND);

    if ( mode & EXECUTE_MODE_PHYSICS )
        aircraft->SetThrottle( mThrottle );
}

void
CommandThrottle::Write( ostream& strm ) const
{
CHECK_TYPESIG(this,TYPESIG_COMMAND);

    strm << COMMAND_NAME_THROTTLE << ' ' << mThrottle << ' ';  // omit endl
}

////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////  CommandSpeed  ///////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/*****************************************************************************
 * Set speed command.
 *****************************************************************************/
CommandSpeed::CommandSpeed( const Speed speed )
:   mSpeed(speed)
{
}

void
CommandSpeed::Execute( shptr<Aircraft> aircraft, const ExecuteMode mode )
{
CHECK_TYPESIG(this,TYPESIG_COMMAND);

    if ( mode & EXECUTE_MODE_PHYSICS )
        aircraft->GetPhysics().SetSpeed( mSpeed );
}

void
CommandSpeed::Write( ostream& strm ) const
{
CHECK_TYPESIG(this,TYPESIG_COMMAND);

    // Record as KPH.
    const fp kph = physics::conv::Speed2KPH(mSpeed).FP();
    strm << COMMAND_NAME_SPEED << ' ' << kph << ' ';  // omit endl
}

////////////////////////////////////////////////////////////////////////////////
///////////////////////////  CommandLandingGear  ///////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/*****************************************************************************
 * Set landing gear command.
 *****************************************************************************/
CommandLandingGear::CommandLandingGear( const bool down )
:   mDown(down)
{
}

void
CommandLandingGear::Execute( shptr<Aircraft> aircraft, const ExecuteMode mode )
{
CHECK_TYPESIG(this,TYPESIG_COMMAND);

    if ( mode & EXECUTE_MODE_PHYSICS )
        aircraft->SetLandingGear( mDown );
}

void
CommandLandingGear::Write( ostream& strm ) const
{
CHECK_TYPESIG(this,TYPESIG_COMMAND);

    strm << COMMAND_NAME_LANDING_GEAR << ' ' << mDown << ' ';  // omit endl
}

////////////////////////////////////////////////////////////////////////////////
/////////////////////////////  CommandBrakes  //////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/*****************************************************************************
 * Set brakes command.
 *****************************************************************************/
CommandBrakes::CommandBrakes( const bool brakes )
:   mBrakes(brakes)
{
}

void
CommandBrakes::Execute( shptr<Aircraft> aircraft, const ExecuteMode mode )
{
CHECK_TYPESIG(this,TYPESIG_COMMAND);

    if ( mode & EXECUTE_MODE_PHYSICS )
        aircraft->EnableBrakes( mBrakes );
}

void
CommandBrakes::Write( ostream& strm ) const
{
CHECK_TYPESIG(this,TYPESIG_COMMAND);

    strm << COMMAND_NAME_BRAKES << ' ' << mBrakes << ' ';  // omit endl
}

} // namespace control
