/*****************************************************************************
 * $LastChangedDate: 2009-11-22 21:39:11 -0600 (Sun, 22 Nov 2009) $
 * @file
 * @author  Jim E. Brooks  http://www.palomino3d.org
 * @brief   Commands to control an Aircraft.
 *//*
 * LEGAL:   COPYRIGHT (C) 2008 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#ifndef CONTROL_COMMAND_HH
#define CONTROL_COMMAND_HH 1

#include "math/module.hh"
#include "math/quaternion.hh"
using namespace math;
#include "object/module.hh"
#include "object/aircraft.hh"
using namespace object;

namespace control {

class Command;

// Asymmetrically, operator<<() (write) is defined, but operator>>() (read) is N/A,
// since reading a derivative from stream into base is illogical.
// Rather, use the factory method Command::Read(istream&).

ostream& operator<<( ostream& strm, const Command& command );

////////////////////////////////////////////////////////////////////////////
/// @brief Defines a command to control an Aircraft.
///
/// This is both an abstract class and a factory class.
/// Read(istream&) can construct a derivative from a stream.
///
class Command : public Shared
{
public:
    ////////////////////////////////////////////////////////////////////////////
    /// @brief Supports graphical or physical replay modes.
    ///
    /// EXECUTE_MODE_PHYSICS is the most realistic but vulnerable
    /// to malfunctioning if system is running too slow or
    /// player selects different aircraft type.
    /// EXECUTE_MODE_GRAPHICS_PHYSICS is a compromise which ensures
    /// every point is replayed while replaying control (gears, brakes, etc).
    ///
    enum ExecuteMode
    {
        // -- update replay.lua --
        EXECUTE_MODE_GRAPHICS           = 1<<0,
        EXECUTE_MODE_PHYSICS            = 1<<1,
        EXECUTE_MODE_GRAPHICS_PHYSICS   = EXECUTE_MODE_GRAPHICS | EXECUTE_MODE_PHYSICS,
        EXECUTE_MODE_ALWAYS             = EXECUTE_MODE_GRAPHICS | EXECUTE_MODE_PHYSICS,
        EXECUTE_MODE_DEFAULT            = EXECUTE_MODE_GRAPHICS_PHYSICS,
    };
                        Command( void ) { SET_TYPESIG(this,TYPESIG_COMMAND); }
    virtual             ~Command()      { INVALIDATE_TYPESIG(this,TYPESIG_COMMAND); }
    virtual void        Execute( shptr<Aircraft> aircraft, const ExecuteMode mode ) = 0;
    virtual void        Write( ostream& strm ) const = 0;
    CLASS_METHOD shptr<Command> Read( istream& strm );  // factory method

    // Unavoidable type codes.
    enum TypeNum { TYPE_NOP, TYPE_POSITION, TYPE_QUATERNION, TYPE_PHYSICAL_ROTATE,
           TYPE_SPEED, TYPE_THROTTLE, TYPE_LANDING_GEAR, TYPE_BRAKES };
    virtual TypeNum     GetType( void ) const = 0;

public:
    DECL_TYPESIG(TYPESIG_COMMAND)
};

////////////////////////////////////////////////////////////////////////////
/// @brief NOP command.
///
/// For Replay::Step default ctor.
///
class CommandNop : public Command
{
public:
                        CommandNop( void ) { }
    TypeNum             GetType( void ) const { return TYPE_NOP; }
    virtual void        Execute( shptr<Aircraft> aircraft, const ExecuteMode mode ) { }
    virtual void        Write( ostream& strm ) const { }
};

////////////////////////////////////////////////////////////////////////////
/// @brief Set position command.
///
class CommandPosition : public Command
{
public:
                        CommandPosition( const GeoVertex& geoPos );
    TypeNum             GetType( void ) const { return TYPE_POSITION; }
    virtual void        Execute( shptr<Aircraft> aircraft, const ExecuteMode mode );
    virtual void        Write( ostream& strm ) const;
    WorldVertex         GetWorldVertex( void ) const;

private:
    const GeoVertex     mGeoPos;
};

////////////////////////////////////////////////////////////////////////////
/// @brief Set graphical rotation (quaternion) command.
///
class CommandQuaternion : public Command
{
public:
                        CommandQuaternion( const Quaternion& quat );
    TypeNum             GetType( void ) const { return TYPE_QUATERNION; }
    virtual void        Execute( shptr<Aircraft> aircraft, const ExecuteMode mode );
    virtual void        Write( ostream& strm ) const;
    Quaternion          GetQuaternion( void ) const { CHECK_TYPESIG(this,TYPESIG_COMMAND);
                                                      return mQuat; }

private:
    const Quaternion    mQuat;
};

////////////////////////////////////////////////////////////////////////////
/// @brief Rotation according to physics model (for physical replay).
/// Correlates to Aircraft::PhysicalRotate().
///
class CommandPhysicalRotate : public Command
{
public:
                        CommandPhysicalRotate( uint axis, fp controlFraction, Milliseconds controlFreq );
    TypeNum             GetType( void ) const { return TYPE_PHYSICAL_ROTATE; }
    virtual void        Execute( shptr<Aircraft> aircraft, const ExecuteMode mode );
    virtual void        Write( ostream& strm ) const;

private:
    const uint          mAxis;
    const fp            mControlFraction;
    const Milliseconds  mControlFreq;
};

////////////////////////////////////////////////////////////////////////////
/// @brief Set speed.
///
class CommandSpeed : public Command
{
public:
                        CommandSpeed( const Speed speed );
    TypeNum             GetType( void ) const { return TYPE_SPEED; }
    virtual void        Execute( shptr<Aircraft> aircraft, const ExecuteMode mode );
    virtual void        Write( ostream& strm ) const;
    Speed               GetSpeed( void ) const { CHECK_TYPESIG(this,TYPESIG_COMMAND);
                                                 return mSpeed; }
private:
    const Speed mSpeed;
};

////////////////////////////////////////////////////////////////////////////
/// @brief Set throttle command.
///
class CommandThrottle : public Command
{
public:
                        CommandThrottle( const fp throttle );
    TypeNum             GetType( void ) const { return TYPE_THROTTLE; }
    virtual void        Execute( shptr<Aircraft> aircraft, const ExecuteMode mode );
    virtual void        Write( ostream& strm ) const;

private:
    const fp mThrottle;
};

////////////////////////////////////////////////////////////////////////////
/// @brief Set landing gear command.
///
class CommandLandingGear : public Command
{
public:
                        CommandLandingGear( const bool down );
    TypeNum             GetType( void ) const { return TYPE_LANDING_GEAR; }
    virtual void        Execute( shptr<Aircraft> aircraft, const ExecuteMode mode );
    virtual void        Write( ostream& strm ) const;

private:
    const bool mDown;
};

////////////////////////////////////////////////////////////////////////////
/// @brief Set brakes command.
///
class CommandBrakes : public Command
{
public:
                        CommandBrakes( const bool brakes );
    TypeNum             GetType( void ) const { return TYPE_BRAKES; }
    virtual void        Execute( shptr<Aircraft> aircraft, const ExecuteMode mode );
    virtual void        Write( ostream& strm ) const;

private:
    const bool mBrakes;
};

////////////////////////////////////////////////////////////////////////////////
/// @brief Exception in Command.
///
class ExceptionCommand : public base::Exception
{
public:
    ExceptionCommand( const string& text ) : Exception(text) { }
};

} // namespace control

#endif // CONTROL_COMMAND_HH
