/*****************************************************************************
 * $LastChangedDate: 2009-11-22 21:39:11 -0600 (Sun, 22 Nov 2009) $
 * @file
 * @author  Jim E. Brooks  http://www.palomino3d.org
 * @brief   Base class for particle-systems.
 *//*
 * LEGAL:   COPYRIGHT (C) 2008 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#define FX_PARTSYS_CC 1
#include "base/module.hh"
using namespace base;
#include "world/module.hh"
using namespace world;
#include "fx/module.hh"
#include "fx/defs.hh"
#include "fx/fx.hh"
#include "fx/partsys.hh"

namespace fx {

////////////////////////////////////////////////////////////////////////////////
//////////////////////////////  ParticleSystem  ////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/*****************************************************************************
 * ctor/dtor.
 *****************************************************************************/
ParticleSystem::ParticleSystem( const Args& args )
:   mTimeOfDeath(base::Timer::GetInstance().GetElapsedTime() + args.mLifetime),
    mTimerFunctor(new ParticleSystem::TickFunctor(*this)),
    mRegisteredTimerFunctor(false),
    mZombie(false)
{
    SET_TYPESIG(this,TYPESIG_PARTICLE_SYSTEM);

    // Enable updating particles.
    if ( args.mTimerEnabled )
    {
        mRegisteredTimerFunctor = true;
        Timer::GetInstance().Register( mTimerFunctor.PTR(), fx::defs::PARTICLE_SYSTEM_TICK_FREQ );
    }
}

ParticleSystem::~ParticleSystem()
{
    // Must unregister!  Else functor would access a destroyed object.
    if ( mRegisteredTimerFunctor )
        Timer::GetInstance().Unregister( mTimerFunctor.PTR() );

    INVALIDATE_TYPESIG(this,TYPESIG_PARTICLE_SYSTEM);
}

/*****************************************************************************
 * Tick.
 *****************************************************************************/
void
ParticleSystem::Tick( const Milliseconds millisecElapsed )
{
CHECK_TYPESIG(this,TYPESIG_PARTICLE_SYSTEM);

    // Time to die?
    if ( UX( millisecElapsed >= mTimeOfDeath ) )
        Expire();  // RIP
}

// Reflect.
void
ParticleSystem::TickFunctor::operator()( const Milliseconds millisecElapsed )
{
CHECK_TYPESIG(&mParticleSystem,TYPESIG_PARTICLE_SYSTEM);

    mParticleSystem.Tick( millisecElapsed );
}

/*****************************************************************************
 * Expire (zombify) particle-system.
 *****************************************************************************/
void
ParticleSystem::Expire( void )
{
CHECK_TYPESIG(this,TYPESIG_PARTICLE_SYSTEM);

    // Turn into a zombie to be destroyed later by the FX singleton.
    if ( not mZombie )  // once
    {
    #if DEBUG_PARTSYS
        CDEBUG << "ParticleSystem turning into zombie" << endl;
    #endif

        // Zombify this.
        mZombie = true;
        GET_FX().AddZombie( this );  // schedule deletion

        // Stop timer-ticks.
        if ( mRegisteredTimerFunctor )
        {
            mRegisteredTimerFunctor = false;
            Timer::GetInstance().Unregister( mTimerFunctor.PTR() );
        }
    }
}

} // namespace fx
