/*****************************************************************************
 * $LastChangedDate: 2009-12-14 19:05:10 -0600 (Mon, 14 Dec 2009) $
 * @file
 * @author  Jim E. Brooks  http://www.palomino3d.org
 * @brief   Missiles.
 *//*
 * LEGAL:   COPYRIGHT (C) 2008 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#define GAME_MISSILE_CC 1
#include "base/module.hh"
#include "base/timer.hh"
using namespace base;
#include "math/module.hh"
#include "math/funcs_trig.hh"
#include "math/funcs_vector.hh"
#include "math/matrix.hh"
#include "math/matrix_funcs.hh"
using namespace math;
#include "graph/module.hh"
#include "graph/model_cache.hh"
#include "graph/scene_graph.hh"
using namespace graph;
#include "object/module.hh"
#include "object/dyna.hh"
using namespace object;
#include "world/module.hh"
using namespace world;
#include "physics/module.hh"
using namespace physics;
#include "control/module.hh"
#include "control/defs_axis.hh"
using namespace control;
#include "collision/module.hh"
#include "collision/events.hh"
using namespace collision;
#include "fx/module.hh"
#include "fx/fx.hh"
using namespace fx;
#include "sound/module.hh"
using namespace sound;
#include "game/module.hh"
#include "game/missile.hh"
#include "game/defs.hh"

namespace game {

INTERN const Meter        MISSILE_LENGTH_GEO( 3.0f );
INTERN const Meter        MISSILE_HIT_DISTANCE_GEO( 15.0f );
INTERN const Meter        MISSILE_SAFE_DISTANCE_GEO( 100.0f );
INTERN const Meter        MISSILE_MAX_DISTANCE_GEO( 4500.0f );  // phony
INTERN const string       MISSILE_MODEL_FILENAME = "models/misc/aim9.ac.gz";
INTERN const fp           MISSILE_SCALE = 1.0f;
INTERN const Meter        MISSILE_EXPLOSION_RADIUS_GEO( 40.0f );
INTERN const Milliseconds MISSILE_EXPLOSION_LIFETIME( 30*1000 );
INTERN const Radian       MANEUVER_ANGLE_ROLL(  Deg2Rad( 10.0f ) );  // per tick
INTERN const Radian       MANEUVER_ANGLE_PITCH( Deg2Rad( 3.0f ) );  // per tick

////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////  Missile  ///////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/*****************************************************************************
 * ctor/dtor.
 *****************************************************************************/
Missile::Missile( shptr<Object> target,
                  Object& launcher,
                  const Speed launcherSpeed,
                  const Meter particleRadiusGeo,
                  const bool malfunction )
:   // Graphical:
    // Temporarily make missile indestructible
    // else it will collide immediately into launcher!
    Dyna(MakeGraph(),launcher.GetPosition(),NON_COLLIDABLE),
    mTarget(target),

    // Physical:
    // Launch missile fast (if player is slow) but always faster than player.
    mSpeedKPH( MAX( 1500.0f, 2.0f * physics::conv::Speed2KPH(launcherSpeed).FP() ) ),
    mMaxDistance(world::conv::Meters2Sim(MISSILE_MAX_DISTANCE_GEO)),
    mInitialPosition(launcher.GetPosition()),

    // Guidance:
    mDirection(0,0,0),  // zeros ok
    mRotationAxis(AXIS_ROLL),
    mManeuverAngleRoll(MANEUVER_ANGLE_ROLL),
    mManeuverAnglePitch(MANEUVER_ANGLE_PITCH),
    mPredictedManeuverAngleSign(1.0f),

    // Animation:
    mPrevTick(0),
    mLauncherPosition(launcher.GetPosition()),
    mCollidable(false),
    mHit(false),
    mHitDistance(world::conv::Meters2Sim(MISSILE_HIT_DISTANCE_GEO)),
    mMalfunction(malfunction)
{
    SET_TYPESIG(this,TYPESIG_MISSILE);

    // Attach missile.
    GET_SCENE_GRAPH().AttachObject( this );  // holds a reference

    // Missile inherits matrix from its launcher.
    SetMatrix( launcher.GetMatrix() );

    // Create particle system for smoke trail.
    GET_FX().MakeSmokeTrail( Milliseconds(30*1000), this );
}

shptr<Graph>
Missile::MakeGraph( void )
{
    // Load and orient 3D model of missile.
    RotationList rotationList;
    rotationList.AddRotation( RotationList::Rotation( YY, Degree(-90.0f) ) );
    return ModelCache::GetInstance().LoadModel( MISSILE_MODEL_FILENAME,
                                                MISSILE_LENGTH_GEO,
                                                GeoVertex(),
                                                rotationList,
                                                MISSILE_SCALE,
                                                ModelCache::LOAD_COPY );
}

Missile::~Missile()
{
    // Don't re-detach from World (illogical as World would hold
    // a reference so ~Missile() would never be called).

    INVALIDATE_TYPESIG(this,TYPESIG_MISSILE);
}

/*****************************************************************************
 * Animate the missile.
 *****************************************************************************/
void
Missile::Tick( const Milliseconds millisecElapsed )
{
CHECK_TYPESIG(this,TYPESIG_MISSILE);

    // Has missile flown too far?
    const fp distanceFromInitialPosition = Distance( GetPosition(), mInitialPosition );
    if ( distanceFromInitialPosition > mMaxDistance )
    {
        Zombify();  // ok to call repeatedly
        return;
    }

    // Wait until start/end of time segment is known.
    if ( mPrevTick > Milliseconds(0) )
    {
        // Compute step (distance in world space).
        const Milliseconds timeDelta = millisecElapsed - mPrevTick;
        const fp step1 = -physics::conv::KPH2Step( timeDelta, mSpeedKPH );  // 1D and negative

        // Once missile is safely away from launcher, enable collision-detection.
        // Another use is to let missile fly straight for a moment before guidance maneuvers it.
        const bool safelyAway = Distance( GetPosition(), mLauncherPosition )
                              > world::conv::Meters2Sim( MISSILE_SAFE_DISTANCE_GEO );

        if ( EX( safelyAway and not mMalfunction ) )
        {
            // The missile-guidance method recomputes matrix
            // and returns a step vector.
            const WorldVertex step3 = Guide( step1 );  // 3D
            const WorldVertex pos = GetPosition() + step3;
            SetPosition( pos );
        }
        else
        {
            // Fly straight aimlessly.
            Translate( ZZ, step1 );
        }

        // Make missile collidable after it has flown away from launcher.
        if ( not mCollidable and safelyAway )
        {
            mCollidable = true;
            SetCollidable( true );
        }
    }

    // Remember tick to measure time-segment.
    mPrevTick = millisecElapsed;
}

/*****************************************************************************
 * @return Forward displacement (step) of missile to move in this frame/tick.
 *****************************************************************************/
WorldVertex
Missile::Guide( const fp step1 )
{
CHECK_TYPESIG(this,TYPESIG_MISSILE);
CHECK_TYPESIG(mTarget,TYPESIG_OBJECT);

    Matrix            missileMatrix = GetMatrix();
    const WorldVertex missilePos    = missileMatrix.GetPosition();
    const WorldVertex targetPos     = mTarget->GetPosition();

    // Hit target?
    if ( mHit )  // safeguard
    {
        return WorldVertex(0,0,0);
    }
    else if ( Distance( targetPos, missilePos ) < mHitDistance
           or IfCollision() )
    {
        mHit = true;
        GET_SOUND().Play( "hit.wav.gz", sound::defs::PLAY_ONCE );

        // Collision has happened.
        SetCollision( Object::COLLISION_FATAL );
        mTarget->SetCollision( Object::COLLISION_FATAL );

        // Broadcast collision event (even though this isn't part of the collision module).
        EVENT_COLLISION.Broadcast( make_pair< shptr<Object>, shptr<Object> >( this, mTarget.PTR() ) );

        // Schedule deletion of missile.
        Zombify();

        // Explosion!
        GET_FX().MakeExplosion( MISSILE_EXPLOSION_LIFETIME,
                                GetPosition(),
                                world::conv::Meters2Sim( MISSILE_EXPLOSION_RADIUS_GEO ) );

        // Stop moving.
        return WorldVertex(0,0,0);
    }

    // Roll or pitch missile.
    switch ( mRotationAxis )
    {
        case AXIS_ROLL:
        {
            Maneuver( missileMatrix,  // IN/OUT
                      AXIS_ROLL, AXIS_PITCH,
                      mManeuverAngleRoll,
                      targetPos );
        }
        break;

        case AXIS_PITCH:
        {
            Maneuver( missileMatrix,  // IN/OUT
                      AXIS_PITCH, AXIS_ROLL,
                      mManeuverAnglePitch,
                      targetPos );
        }
        break;
        
        default:
        break;  // else continue same direction
    }

    // Compute direction and save it in case later missile flies straight aimlessly.
    mDirection[XX] = missileMatrix[Xz] * step1;
    mDirection[YY] = missileMatrix[Yz] * step1;
    mDirection[ZZ] = missileMatrix[Zz] * step1;

    return mDirection;
}

/*****************************************************************************
 * subroutine
 * This maneuvers the missile by recomputing its matrix orientation. 
 *****************************************************************************/
void
Missile::Maneuver( Matrix& matrix,  /*IN/OUT*/
                   const uint rotationAxis,
                   const uint nextRotationAxis,
                   Radian maneuverAngle, 
                   const WorldVertex targetPos )
{
CHECK_TYPESIG(this,TYPESIG_MISSILE);
CHECK_TYPESIG(mTarget,TYPESIG_OBJECT);
ASSERT_AXIS(rotationAxis);
ASSERT_AXIS(nextRotationAxis);

//..............................................................................
// Macros.

// Compute matrix to transform a world vertex into missile's local space.
#define MISSILE_TRANSPOSE_MATRIX() \
    MatrixTranspose( transposeMatrix, matrix );

// Transpose target's position into missile's local space.
#define COMPUTE_TRANSPOSE_TARGET_POS() \
    TranslateRotate<Vector3,WorldVertex,Matrix>( targetPos, transposeMatrix );

// Compute angle between 2 axises of missile (in its local space)
// and the target position (transposed into missile's local space).
#define COMPUTE_ANGLE_MISSILE_AND_TARGET()                                                                      \
    (rotationAxis == AXIS_ROLL) ? Angle3( Vector3( transposeTargetPos[XX], transposeTargetPos[YY], 0.0f ),     \
                                          Vector3( 0.0f, 1.0f, 0.0f  ) )                                       \
    /* AXIS_PITCH */            : Angle3( Vector3( 0.0f, transposeTargetPos[YY], transposeTargetPos[ZZ] ),     \
                                          Vector3( 0.0f, 0.0f, -1.0f ) );

//..............................................................................

    // Transform target into missile's local space.
    Matrix transposeMatrix;
    MISSILE_TRANSPOSE_MATRIX();
    Vector3 transposeTargetPos = COMPUTE_TRANSPOSE_TARGET_POS();

    // Compute angle between an axis of missile and target.
    Radian angle = COMPUTE_ANGLE_MISSILE_AND_TARGET();

    // Use maximum maneuver angle or actual angle if actual angle is smaller.
    // That is, maneuver angle will narrow.
    if ( ABS(angle) < maneuverAngle )
    {
        // This axis of rotation has narrowed as far as it can go.
        // Switch state to maneuver in another axis-of-rotation.
        maneuverAngle = angle;
        mRotationAxis = nextRotationAxis;  // next state
    }

    // Optimize by rotating the same direction as last time
    // to avoid reverse-rotation case.
    maneuverAngle = Radian(fp(maneuverAngle) * mPredictedManeuverAngleSign);
    const Radian minAngle = Deg2Rad( 0.0025f );
    if ( ABS(fp(angle)) > fp(minAngle) )
    {
        // Rotate in one direction.  If the angle increased by mistake,
        // then rotate in the other direction (tracking is done by trial-and-error).
        MatrixRotateLocal( matrix, rotationAxis, maneuverAngle );

        // Update after recomputing matrix.
        MISSILE_TRANSPOSE_MATRIX();
        transposeTargetPos = COMPUTE_TRANSPOSE_TARGET_POS();

        Radian angle2 = COMPUTE_ANGLE_MISSILE_AND_TARGET();
        if ( ABS(fp(angle2)) > ABS(fp(angle)) )
        {
            mPredictedManeuverAngleSign *= -1.0f;
            MatrixRotateLocal( matrix, rotationAxis, -maneuverAngle * 2.0f );
        }
    }

    // FYI: This will cause spiraling as if missile was malfunctioning.
  //mRotationAxis = nextRotationAxis;  // next state

    // Commit rotations.
    SetMatrix( matrix );
}

} // namespace game
