/*****************************************************************************
 * $LastChangedDate: 2009-11-22 21:39:11 -0600 (Sun, 22 Nov 2009) $
 * @file
 * @author  Jim E. Brooks  http://www.palomino3d.org
 * @brief   Missiles.
 *//*
 * LEGAL:   COPYRIGHT (C) 2008 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#ifndef GAME_MISSILE_HH
#define GAME_MISSILE_HH 1

#include "object/module.hh"
#include "object/dyna.hh"
using namespace object;
#include "physics/module.hh"
using namespace physics;

namespace game {

////////////////////////////////////////////////////////////////////////////////
/// @brief Missiles.
///
/// Lua:
/// ----
/// Lua determines which Actor is the target.
/// Lua interfaces with this C++ class in terms of the Dyna class.
///
/// C++:
/// ----
/// Missile inherits its matrix and speed from its launcher.
/// Every tick, a guidance formula rotates the Dyna's matrix
/// to guide the missile at the target.
/// If the target is hit, movement stops.
/// Missile does its own distance-based collision-detection.
/// If the target is destroyed by another action,
/// missile will fly straight indefinitely.
///
class Missile : public Dyna
{
PREVENT_COPYING(Missile)
public:
                        Missile( shptr<Object> target,
                                 Object& launcher,
                                 const Speed launcherSpeed,
                                 const Meter particleRadiusGeo,
                                 const bool malfunction );
                        ~Missile();
    virtual const string GetName( void ) { return string("Missile"); }

private:
    shptr<Graph>        MakeGraph( void );
    void                Tick( const Milliseconds millisecElapsed );
    WorldVertex         Guide( const fp step1 );
    void                Maneuver( Matrix& matrix,  /*IN/OUT*/
                                  const uint rotationAxis,
                                  const uint nextRotationAxis,
                                  Radian maneuverAngle, 
                                  const WorldVertex targetPos );
private:
    // Graphical:
    shptr<Object>       mTarget;

    // Physical:
    SpeedKPH            mSpeedKPH;          ///< scalar speed
    const fp            mMaxDistance;       ///< prevent missile flying infinitely
    const WorldVertex   mInitialPosition;

    // Guidance:
    WorldVertex         mDirection;         ///< forward step
    uint                mRotationAxis;      ///< AXIS_ROLL or AXIS_PITCH
    Radian              mManeuverAngleRoll;
    Radian              mManeuverAnglePitch;
    fp                  mPredictedManeuverAngleSign;    ///< try rotating in same direction as previously

    // Animation:
    Milliseconds        mPrevTick;
    const WorldVertex   mLauncherPosition;
    bool                mCollidable;        ///< temporarily indestructible while launching
    bool                mHit;               ///< missile is zombified after hit
    const fp            mHitDistance;       ///< scalar in world space
    const bool          mMalfunction;       ///< missiles don't always work
public:
    DECL_TYPESIG(TYPESIG_MISSILE)
};

} // namespace game

#endif // GAME_MISSILE_HH
