/*****************************************************************************
 * $LastChangedDate: 2010-02-05 15:06:48 -0600 (Fri, 05 Feb 2010) $
 * @file
 * @author  Jim E. Brooks  http://www.palomino3d.org
 * @brief   Module configuration.
 *//*
 * LEGAL:   COPYRIGHT (C) 2009 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#define GRAPH_CONF_CC 1
#include "base/module.hh"
#include "base/conf.hh"
#include "base/misc.hh"
using namespace base;
#include "gfx/rgba.hh"
using namespace gfx;
#include "graph/module.hh"
#include "graph/defs.hh"
#include "graph/conf.hh"
#include "graph/shadow_scene_graph.hh"
#include <fstream>

namespace graph {

////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////  Conf  ///////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/*****************************************************************************
 * ctor/dtor.
 *****************************************************************************/
Conf::Conf( void )
:   mValid(false),
    // cmd-line args
    mEnableOsgOptimizer(true),
    mEnableWireframe(false),
    mEnableDumpFull(false),
    mEnableFog(true),
    mShadowMode(graph::defs::SHADOW_MODE_DEFAULT),
    // .conf file
    mShadowLightNum(1),
    mShadowLightPos(500.0f, 500.0f, 3000.0f, 0.0f),
    mShadowLightDir(0.70710678f, 0.0f, -0.70710678f),
    mShadowLightColor(1.0f, 1.0f, 1.0f, 1.0f),
    mBackgroundColor(graph::defs::DEFAULT_BACKGROUND_COLOR)
{
    // NOP
}

/*****************************************************************************
 * Initialize/read configuration.
 *****************************************************************************/
void
Conf::Init( int argc, char** argv, const string& confTxtFilename )
{

BEGIN_ONCE {

    mValid = true;  // assign this first as *_CONF may expand to GetConf() which asserts it

    ParseCmdlineArgs( argc, argv );   // cmdline args parsed prior to reading .conf file
    ReadConfTxt( confTxtFilename );   // read from textual .conf file
    ParseCmdlineArgs2( argc, argv );  // cmdline args that override .conf file

} END_ONCE

}

/*****************************************************************************
 * Parse cmdline args.
 *****************************************************************************/
void
Conf::ParseCmdlineArgs( int argc, char** argv )
{
    // NOP
}

void
Conf::ParseCmdlineArgs2( int argc, char** argv )
{
    for ( int i = 1; i < argc; ++i )
    {
        if ( STREQ( argv[i], "-no-osg-opt" ) )
        {
            mEnableOsgOptimizer = false;
            CmdargDone( i );
            CLOG << "Disabled OSG optimizer." << endl;
        }
        else if ( STREQ( argv[i], "-wire" )
               or STREQ( argv[i], "-wireframe" ) )
        {
            mEnableWireframe = true;
            CLOG << "Enabled wireframe mode." << endl;
        }
        else if ( STREQ( argv[i], "-dump-full" )   // BTW, dumping core is -debug/-debugger/-gdb
               or STREQ( argv[i], "-full-dump" )   // alt
               or STREQ( argv[i], "-dump-all" ) )  // alt
        {
            mEnableDumpFull = true;
            CmdargDone( i );
            CLOG << "Enabled fully dumping scene-graph." << endl;
        }
        else if ( STRSTR( argv[i], "-no-shadow" )  // or -no-shadows
               or STREQ( argv[i], "-fast" ) )
        {
            mShadowMode = graph::defs::SHADOW_MODE_DISABLED;
            CmdargDone( i );
            CLOG << "Disabled shadows." << endl;
        }
        else if ( STRSTR( argv[i], "-fast-shadow" ) )  // or -fast-shadows
        {
            mShadowMode = graph::defs::SHADOW_MODE_SAMPLE1;
            CmdargDone( i );
            CLOG << "Enabled fast shadows." << endl;
        }
        else if ( STRSTR( argv[i], "-soft-shadow" ) )  // or -soft-shadows
        {
            mShadowMode = graph::defs::SHADOW_MODE_SAMPLE4;
            CmdargDone( i );
            CLOG << "Enabled soft shadows." << endl;
        }
    }
}

/*****************************************************************************
 * Read from .conf file.
 *****************************************************************************/
void
Conf::ReadConfTxt( const string& confTxtFilename )
{
    try
    {
        ModuleConfTxt confTxt;
        confTxt.Read( string(base::GetDataDir()) + string(confTxtFilename) );
    }
    catch (...)
    {
        CERROR << "ERROR: cannot read " << confTxtFilename << endl;
        EXIT( 1 );
    }
}

void
Conf::ModuleConfTxt::ProcessField( string& field, ifstream& conf )
{
    bool error = false;

    if ( field == "ShadowLightNum" )
    {
        conf >> GRAPH_CONF.mShadowLightNum;
    }
    else if ( field == "ShadowLightPosition" )  // 4D
    {
        conf >> GRAPH_CONF.mShadowLightPos[XX];
        conf >> GRAPH_CONF.mShadowLightPos[YY];
        conf >> GRAPH_CONF.mShadowLightPos[ZZ];
        conf >> GRAPH_CONF.mShadowLightPos[WW];
    }
    else if ( field == "ShadowLightDirection" )  // 3D
    {
        conf >> GRAPH_CONF.mShadowLightDir[XX];
        conf >> GRAPH_CONF.mShadowLightDir[YY];
        conf >> GRAPH_CONF.mShadowLightDir[ZZ];
    }
    else if ( field == "ShadowLightColor" )
    {
        fp r, g, b, a;
        conf >> r;
        conf >> g;
        conf >> b;
        conf >> a;
        GRAPH_CONF.mShadowLightColor.Set( r, g, b, a );
    }
    else if ( field == "BackgroundColor" )
    {
        fp r, g, b, a;
        conf >> r;
        conf >> g;
        conf >> b;
        conf >> a;
        GRAPH_CONF.mBackgroundColor.Set( r, g, b, a );
    }

    // Errors?
    if ( error )
    {
        CERROR << "ERROR[conf]: parameter for " << field << " is invalid." << endl;
        EXIT( 1 );
    }
}

/*****************************************************************************
 * Print Conf configuration.
 *****************************************************************************/
ostream&
operator<<( ostream& strm, const Conf& conf )
{
    std::ios::fmtflags savedFlags = strm.flags();

    strm << "conf [graph]:"                                                      << endl
         << "enable OSG optimizer        = " << conf.mEnableOsgOptimizer         << endl
         << "shadow mode                 = " << conf.mShadowMode              ;//<< endl
    strm.flags(savedFlags);

    return strm;
}

} // namespace graph
