/*****************************************************************************
 * $LastChangedDate: 2009-11-22 21:39:11 -0600 (Sun, 22 Nov 2009) $
 * @file
 * @author  Jim E. Brooks  http://www.palomino3d.org
 * @brief   Math functions for computer graphics.
 * @remarks - To speed or improve FP execution:
 *            - use multiplication instead of division
 *            - power of 2 operations involve no loss of accuracy
 *            - math::PI is preferred over M_PI etc from math.h.
 *              math.h has confusing #defines such as M_2_PI
 *              which means 2/PI (not 2*PI).
 *          - NaN are caused by division and some standard C math functions.
 * REMINDERS: - Use MATH_EXCEPTION() instead of ASSERT().
 *            - C++ <cmath> is better than C <math.h> because C++ functions are overloaded.
 *//*
 * LEGAL:   COPYRIGHT (C) 2004 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#ifndef MATH_FUNCS_HH
#define MATH_FUNCS_HH 1

#if _MSC_VER
#include <float.h>  // _finite()
#endif // _MSC_VER

#include "math/funcs_debug.hh"

namespace math {

////////////////////////////////////////////////////////////////////////////////
///////////////////////////  C O M P A R I S O N  //////////////////////////////
////////////////////////////////////////////////////////////////////////////////

// IF_NAN() also returns true if infinity
// (divide by zero is technically infinity, not NaN, but IF_NAN() should catch divide errors).
// C99 isfinite() is true if neither NaN nor infinity.
// A NaN never equals anything -- not even another NaN.
// NOTE: !(n==n) is unreliable because the compiler could optimize as always false
// (GNU gcc 3.4 will do that).
#if _MSC_VER
#   define IF_NAN(N) (UX(!_finite(N)))
#else
#   define IF_NAN(N) (UX(!std::isfinite(N)))
#endif

/// Cope with floating-point inaccuracy.
/// Return true if two floats are approximately equal.
template<typename FP> bool FP_EQ( FP a, FP b ) { return ABS(a-b) < 0.001f; }
template<typename FP> bool FP_LE( FP a, FP b ) { return a-0.001f <= b or a+0.001f <= b; }
template<typename FP> bool FP_GE( FP a, FP b ) { return a-0.001f >= b or a+0.001f >= b; }

/// Compare floats: compare full integral part, compare fracDigits of fractional part.
/// Eg: FP_EQ( 1.234, 1.234, 4 )
/// Note: In practice, usually, fracDigits should be small, or miscompares will be frequent.
/// Tested by tests/math.cc.
template<typename FP>
bool
FP_EQ( const FP a, const FP b, const uint fracDigits )
{
    const FP diff = a - b;
    const FP scale = pow( double(10.0), double(fracDigits) );
    const uint64 tolerance = uint64(scale * 0.005f);
    return uint64(ABS(diff * scale)) <= tolerance;
}

/*****************************************************************************
 * @return True If two numbers have the same sign.
 *****************************************************************************/
template<typename T1,typename T2>
bool
IfSameSign( const T1 a, const T2 b )
{
  //return (a * b) >= 0;  // wrong because if either is zero: -n * 0 = 0

    if ( a >= 0 and b >= 0 )  // both positive?
        return true;
    else if ( a < 0 and b < 0 )  // both negative
        return true;
    else
        return false;  // different signs
}

////////////////////////////////////////////////////////////////////////////////
///////////////////////////////  B A S I C   ///////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/*****************************************************************************
 * Select square root function according to type.
 *****************************************************************************/
INLINE float
SQRT( float f )
{
MATH_EXCEPTION( f >= 0.0f, "math error: sqrt(-n) undefined" );
//  return sqrtf( f );
    return std::sqrt( f );  // C++ <cmath>
}

INLINE double
SQRT( double f )  // use actual types (fpx may be double or long double)
{
MATH_EXCEPTION( f >= 0.0f, "math error: sqrt(-n) undefined" );
//  return sqrt( f );
    return std::sqrt( f );  // C++ <cmath>
}

INLINE long double
SQRT( long double f )
{
MATH_EXCEPTION( f >= 0.0f, "math error: sqrt(-n) undefined" );
//  return sqrtl( f );      // FreeBSD omits sqrtl()
    return std::sqrt( f );  // C++ <cmath>
}

////////////////////////////////////////////////////////////////////////////////
////////////////////////////  M I S C   M A T H   //////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/*****************************************************************************
 * x^2
 *****************************************************************************/
template<typename T>
T
SQUARE( T x )
{
    return x * x;
}

/*****************************************************************************
 * Average.
 *****************************************************************************/
template<typename T>
T
AVG( T x, T y )
{
    return (x + y) / 2;
}

/*****************************************************************************
 * Force a number to be within a range.
 *****************************************************************************/
template<typename T>
T
Range( T val, T lo, T hi )
{
    if ( val < lo ) return lo;
    if ( val > hi ) return hi;
    return val;
}

/*****************************************************************************
 * @return True if two lines overlap.
 *****************************************************************************/
template<typename T>
T
IfInRange( T a0, T a1, T b0, T b1 )
{
    // Prove the reverse: test if out-of-range.
         if ( a0 > b1 )
        return false;
    else if ( a1 < b0 )
        return false;
    else if ( b0 > a1 )
        return false;
    else if ( b1 < a0 )
        return false;
    else
        return true;  // within range
}

/*****************************************************************************
 * Compute a fraction for a number in a range.
 * Number is allowed to be outside the range (result will be clamped to 0 or 1).
 * Examples:
 * FractionInRange(100,100,200) = 0.0
 * FractionInRange(150,100,200) = 0.5
 * FractionInRange(200,100,200) = 1.0
 *****************************************************************************/
INLINE fp
FractionInRange( fp n, fp low, fp high )
{
// Negative range undefined.
MATH_EXCEPTION( (low < high) and (low >= 0 and high > 0),
                "math error: FractionInRange()" );

    if ( n <= low )
    {
        return 0.0f;
    }
    else if ( n >= high )
    {
        return 1.0f;
    }
    else  // inside range
    {
        return (n - low) / (high - low);
    }
}

/*****************************************************************************
 * Map a number from one range into another.
 *****************************************************************************/
INLINE fp
Remap( fp value, fp rangeOld, fp rangeNew )
{
    const fp n = value / rangeOld * rangeNew;
MATH_EXCEPTION( not IF_NAN(n), "math error: Remap()" );
    return n;
}

/*****************************************************************************
 * Truncate a float at an interval.
 * Eg, Truncate(24,10) returns 20.
 *****************************************************************************/
INLINE fp
Truncate( fp val, fp interval )
{
    float integral;  // float, not fp, as address is passed to modff()
    std::modf( val/interval, &integral );
    return integral * interval;
}

/*****************************************************************************
 * Modulo that handles negative values by subtracting from limit.
 * For example, ModuloAbs(deg,360.0f) can be used to accomodate a function
 * that requires degrees to be in range {0,..,360}
 * ModuloAbs( -1.0,  360.0 ) = 359.0
 * ModuloAbs( -90.0, 360.0 ) = 270.0
 *****************************************************************************/
template<typename T>
T
ModuloAbs( const T n, const T limit )
{
    const T m = std::fmod( n, limit );
    if ( m >= 0 )
        return m;
    else  // m < 0
        return limit + m;  // effectively subtraction
}

} // namespace math

#endif // MATH_FUNCS_HH
