/*****************************************************************************
 * $LastChangedDate: 2009-11-22 21:39:11 -0600 (Sun, 22 Nov 2009) $
 * @file
 * @author  Jim E. Brooks  http://www.palomino3d.org
 * @brief   Math functions for computer graphics.
 * @remarks - To speed or improve FP execution:
 *            - use multiplication instead of division
 *            - power of 2 operations involve no loss of accuracy
 *            - math::PI is preferred over M_PI etc from math.h.
 *              math.h has confusing #defines such as M_2_PI
 *              which means 2/PI (not 2*PI).
 *          - NaN are caused by division and some standard C math functions.
 * REMINDERS: - Use MATH_EXCEPTION() instead of ASSERT().
 *            - C++ <cmath> is better than C <math.h> because C++ functions are overloaded.
 *//*
 * LEGAL:   COPYRIGHT (C) 2004 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#ifndef MATH_FUNCS_DEBUG_HH
#define MATH_FUNCS_DEBUG_HH 1

namespace math {

////////////////////////////////////////////////////////////////////////////////
///////////////////////////////  D E B U G  ////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

// Same as ASSERT() but throws C++ exception.
// Clients of math now depend on C++ exception.
#define MATH_EXCEPTION( COND, TEXT ) {{ if ( UX( not (COND) ) ) throw Exception( TEXT ); }}

// Latitude should be in {0',..,180.0'}
// Longitude should be in {0',..,360.0'}
#if DEBUG
#define ASSERT_LAT_LON_DEG( LAT, LON )                                  \
    MATH_EXCEPTION( FP_GE<fp>(LAT,0.0f) and FP_LE<fp>(LAT,180.0f)       \
                and FP_GE<fp>(LON,0.0f) and FP_LE<fp>(LON,360.0f),      \
                     "(lat,lon) out-of-range" )
#else
#define ASSERT_LAT_LON_DEG( LAT, LON )  // ok as NOP for RELEASE
#endif

/// True if axis is XX,YY,ZZ.
#define ASSERT_AXIS( axis ) ASSERT( uint(axis) < 3 )

} // namespace math

#endif // MATH_FUNCS_DEBUG_HH
