/*****************************************************************************
 * $LastChangedDate: 2010-01-02 10:38:46 -0600 (Sat, 02 Jan 2010) $
 * @file
 * @author  Jim E. Brooks  http://www.palomino3d.org
 * @brief   Module configuration.
 *
 * Another subtle initiatialization problem:
 *
 * This was unusually put here in conf.cc instead of defs.hh.
 * namespace shader {
 * namespace defs {
 * osg::Vec3 DEFAULT_LIGHT0_POSITION(x,y,z);
 * } // namespace defs
 * } // namespace shader
 *
 * Conf::Conf( void )
 *    mLight0Position(shader::defs::DEFAULT_LIGHT0_POSITION)
 *
 * DEFAULT_LIGHT_POSITION might be initialized AFTER the Conf ctor executes!
 * So try to write only basic constants in defs if possible.
 *
 *//*
 * LEGAL:   COPYRIGHT (C) 2009 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#define SHADER_CONF_CC 1
#include "base/module.hh"
#include "base/conf.hh"
#include "base/misc.hh"
using namespace base;
#include "math/module.hh"
#include "math/defs_vector.hh"
using namespace math;
#include "shader/module.hh"
#include "shader/conf.hh"
#include "shader/defs.hh"
#include <fstream>
#include <osg/Fog>

namespace shader {

////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////  Conf  ///////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/*****************************************************************************
 * ctor/dtor.
 *****************************************************************************/
Conf::Conf( void )
:   mValid(false),
    // cmd-line args
    mEnableShaders(true),
    mEnableLighting(true),
    // .conf file
    mLight0Position(8000.0f, 10000.0f, 3000.0f),  // constants here solve
    mEnableFog(true),
    mFogColor(0.65f, 0.65f, 0.85f, 1.0f),         // initialization problem (see above)
    mFogDensity(0.0007),
    mFogMode(osg::Fog::EXP2),
    mFogMin(0.0f),
    mBrightness(1.0f)
    // Environment:
{
    // NOP
}

/*****************************************************************************
 * Initialize/read configuration.
 *****************************************************************************/
void
Conf::Init( int argc, char** argv, const string& confTxtFilename )
{

BEGIN_ONCE {

    mValid = true;  // assign this first as *_CONF may expand to GetConf() which asserts it

    ParseCmdlineArgs( argc, argv );   // cmdline args parsed prior to reading .conf file
    ReadConfTxt( confTxtFilename );   // read from textual .conf file
    ParseCmdlineArgs2( argc, argv );  // cmdline args that override .conf file

} END_ONCE

}

/*****************************************************************************
 * Parse cmdline args.
 *****************************************************************************/
void
Conf::ParseCmdlineArgs( int argc, char** argv )
{
    // NOP
}

void
Conf::ParseCmdlineArgs2( int argc, char** argv )
{
    for ( int i = 1; i < argc; ++i )
    {
        if ( STRSTR( argv[i], "-no-shader" ) )  // or -no-shaders
        {
            // Disable shaders.
            mEnableShaders = false;
            CLOG << "Disabled shaders." << endl;
        }
        else if ( STRSTR( argv[i], "-no-light" ) )  // or -no-lights
        {
            // Disable lighting.
            mEnableLighting = false;
            CLOG << "Disabled lighting." << endl;
        }
        else if ( STREQ( argv[i], "-no-fog" ) )
        {
            mEnableFog = false;
            CLOG << "Disabled fog." << endl;
        }
        else if ( STREQ( argv[i], "-brightness" ) )
        {
            mBrightness = Argf( i+1, argc, argv );
        }
    }
}

/*****************************************************************************
 * Read from .conf file.
 *****************************************************************************/
void
Conf::ReadConfTxt( const string& confTxtFilename )
{
    try
    {
        ModuleConfTxt confTxt;
        confTxt.Read( string(base::GetDataDir()) + string(confTxtFilename) );
    }
    catch (...)
    {
        CERROR << "ERROR: cannot read " << confTxtFilename << endl;
        EXIT( 1 );
    }
}

void
Conf::ModuleConfTxt::ProcessField( string& field, ifstream& conf )
{
    bool error = false;

    //---------------------  light  ---------------------------------------------
    if ( field == "Light0Position" )
    {
        conf >> SHADER_CONF.mLight0Position[XX];
        conf >> SHADER_CONF.mLight0Position[YY];
        conf >> SHADER_CONF.mLight0Position[ZZ];
    }
    //---------------------  fog  ---------------------------------------------
    else if ( field == "FogColor" )
    {
        fp r, g, b, a;
        conf >> r;
        conf >> g;
        conf >> b;
        conf >> a;
        SHADER_CONF.mFogColor.Set( r, g, b, a );
    }
    else if ( field == "FogMode" )
    {
        string s; conf >> s;  // string
        if ( s == "linear" )
        {
            SHADER_CONF.mFogMode = osg::Fog::LINEAR;
        }
        else if ( s == "exp2" )
        {
            SHADER_CONF.mFogMode = osg::Fog::EXP2;
        }
        else  // default
        {
            error = (s != "exp");
            SHADER_CONF.mFogMode = osg::Fog::EXP;
        }
    }
    else if ( field == "FogDensity" )
    {
        conf >> SHADER_CONF.mFogDensity;
    }
    else if ( field == "FogMin" )
    {
        conf >> SHADER_CONF.mFogMin;
    }
    
    // Errors?
    if ( error )
    {
        CERROR << "ERROR[conf]: parameter for " << field << " is invalid." << endl;
        EXIT( 1 );
    }
}

/*****************************************************************************
 * Print Conf configuration.
 *****************************************************************************/
ostream&
operator<<( ostream& strm, const Conf& conf )
{
    std::ios::fmtflags savedFlags = strm.flags();

    strm << "conf [shader]:"                                                      << endl
         << "enable shaders              = " << conf.mEnableShaders            ;//<< endl
    strm.flags(savedFlags);

    return strm;
}

} // namespace shader
