/*****************************************************************************
 * $LastChangedDate: 2010-02-03 18:48:51 -0600 (Wed, 03 Feb 2010) $
 * @file
 * @author  Jim E. Brooks  http://www.palomino3d.org
 * @brief   Window class.
 *//*
 * LEGAL:   COPYRIGHT (C) 2007 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#ifndef VIEW_WINDOW_HH
#define VIEW_WINDOW_HH 1

#define GET_WINDOW()        (Window::GetInstance())
#define GET_MAIN_VIEW()     (Window::GetInstance().GetMainView())
#define POST_REDISPLAY()    // NOP

#include "gfx/graphics.hh"
using namespace gfx;
#include "view/view.hh"

#if ! DOXYGEN
#include <osg/GraphicsContext>
namespace osg { class Object; class Operation; }
namespace osgGA { class GUIEventHandler; }
namespace osgViewer { class Viewer; class GraphicsWindow; class StatsHandler; }
#endif // ! DOXYGEN

namespace view {

class View; class Manipulator;

////////////////////////////////////////////////////////////////////////////////
/// @brief Window contains Views, executes main loop, broadcasts draw events.
///
class Window
{
PREVENT_COPYING(Window)
private:
                            Window( const Vector2& pos, const WidthHeight<int>& wh, const string& title );
                            ~Window();
    CLASS_METHOD Window*    CreateInstance( void );

public:
    CLASS_METHOD Window&    GetInstance( void )
    {
        // NewInstance() gets ctor args from conf.
        if ( EX( msInstance != NULL ) )
            return *msInstance;
        else
            return *(msInstance = Window::CreateInstance());
    }
    void                    Make( void ) { }  // (NOP as GetInstance() creates it)

// The main graphics loop:
public:
    void                    EnterMainLoop( void );

// Views:
public:
    enum { MAIN_VIEW = true, NOT_MAIN_VIEW = false };
    void                    AddView( shptr<View> view, const bool mainView = false );
    shptr<View>             GetMainView( void )
                            {
                            ASSERT( not mViews.empty() );
                                return mViews.front();
                            }
    bool                    SaveViewpoint( void );
    bool                    RestoreViewpoint( void );

// Info:
public:
    WidthHeight<int>        GetWindowSize( void );
    Rect<int>               GetWindowRect( void );
    Rect<int>               GetWindowPosition( void );
    uint                    GetFPS( void ) { return mFPS.ComputeFPS(); }

// OSG (LIMITED USE):
public:
    enum eCamera { eCamera_3D, eCamera_2D };
    void                              OsgAddCamera( RefPtr<osg::Camera> camera, const eCamera cameraType = eCamera_3D, const bool mainView = NOT_MAIN_VIEW );
    RefPtr<osgViewer::GraphicsWindow> OsgGetGraphicsWindow( void );
    void                              OsgRegisterEventHandler( RefPtr<osgGA::GUIEventHandler> eventHandler );
    void                              OsgSetStatsKey( const int key );

private:
    void                    ComputeFPS( const Milliseconds elapsedMilsec );
    void                    DoOperation( void );
    void                    DrawViewOutline( shptr<View> view, const RGBA outlineColor );

private:
    class Operation : public osg::Operation
    {
    public:
        Operation( void );
        virtual ~Operation();
        virtual void operator ()( osg::Object* );
    };

private:
    struct FPS
    {
                        FPS( void );
        void            Tick( const Milliseconds elapsedMilsec );
        uint            ComputeFPS( void ) { return mFPS; }

        uint            mFPS;           ///< FPS of previous second
        uint            mFrameCount;    ///< frame count of current second
        Milliseconds    mFrameStart;    ///< beginning of current second
    };

private:
    typedef list<shptr<View> > Views;
    CLASS_VAR Window*                       msInstance;         ///< there can be only one
    RefPtr<osgViewer::Viewer>               mViewer;            ///< contains cameras
    Views                                   mViews;             ///< Window contains one or more Views
    string                                  mTitle;             ///< window title
    FPS                                     mFPS;               ///< frame rate
    RefPtr<osg::GraphicsContext::Traits>    mTraits;
    RefPtr<osg::GraphicsContext>            mGraphicsContext;
    RefPtr<osgViewer::GraphicsWindow>       mGraphicsWindow;
    RefPtr<osg::Operation>                  mOperation;
    RefPtr0<osgViewer::StatsHandler>        mStatsHandler;      ///< statistics (might be disabled/NULL)
    RefPtr<Manipulator>                     mManipulator;       ///< custom view manipulator (OSG specific)
};

} // namespace view

#endif // VIEW_WINDOW_HH
