/*****************************************************************************
 * $LastChangedDate: 2010-02-14 12:12:43 -0600 (Sun, 14 Feb 2010) $
 * @file
 * @author  Jim E. Brooks  http://www.palomino3d.org
 * @brief   Module configuration.
 *//*
 * LEGAL:   COPYRIGHT (C) 2007 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#define WORLD_CONF_CC 1
#include "base/module.hh"
#include "base/conf.hh"
#include "base/misc.hh"
using namespace base;
#include "math/module.hh"
#include "math/funcs.hh"
using namespace math;
#include "gfx/module.hh"
#include "gfx/rgba.hh"
using namespace gfx;
#include "control/module.hh"
using namespace control;
#include "world/module.hh"
#include "world/conf.hh"

namespace world {

////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////  Conf  ///////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/*****************************************************************************
 * ctor/dtor.
 *****************************************************************************/
Conf::Conf( void )
:   mValid(false),
    // cmd-line args
    mFlatWorld(true),
    mEnableOsgOptimizer(true),
    // .conf file
    // World:
    mObjectScale(1.0),
    mObjectScale3(1.0,1.0,1.0),
    // Terrain model of flat world:
    mFlatModelFile(""),
    mFlatModelRotationList(),
    mFlatModelScale(1.0f),
    mFlatModelOffset(0,0,0),
    mFlatModelNW(Degree(0),Degree(0),Meter(0)),
    mFlatModelSE(Degree(0),Degree(0),Meter(0)),
    mFlatModelMeterOffset(Meter(0),Meter(0),Meter(0)),
    mFlatModelMeterSimRatio(1.0f),
    mFlatModelMeterLowAlt(100.0f),
    mFlatModelMeterAltMin(0.0f),
    mFlatModelMeterAltMax(10000.0f)
{
    // NOP
}

/*****************************************************************************
 * Initialize/read configuration.
 *****************************************************************************/
void
Conf::Init( int argc, char** argv, const string& confTxtFilename )
{

BEGIN_ONCE {

    mValid = true;  // assign this first as *_CONF may expand to GetConf() which asserts it

    ParseCmdlineArgs( argc, argv );   // cmdline args parsed prior to reading .conf file
    ReadConfTxt( confTxtFilename );   // read from textual .conf file
    ParseCmdlineArgs2( argc, argv );  // cmdline args that override .conf file

} END_ONCE

}

/*****************************************************************************
 * Parse cmdline args.
 *****************************************************************************/
void
Conf::ParseCmdlineArgs( int argc, char** argv )
{
    for ( int i = 1; i < argc; ++i )
    {
        if ( STRSTR( argv[i], "-flat" ) )  // -flat-world
        {
            mFlatWorld = true;
            CmdargDone( i );
            CLOG << "Enabled flat world." << endl;
        }
#if COMPILE_OSSIM
        else if ( STRSTR( argv[i], "-spher" ) )  // -sphere or -sphere-world or -spherical
        {
            mFlatWorld = false;
            CLOG << "Enabled spherical world." << endl;
            CmdargDone( i );
        }
#endif
        else if ( STREQ( argv[i], "-no-osg-opt" ) )
        {
            mEnableOsgOptimizer = false;
            CmdargDone( i );
            CLOG << "Disabled OSG optimizer." << endl;
        }
    }
}

void
Conf::ParseCmdlineArgs2( int argc, char** argv )
{
    // NOP
}

/*****************************************************************************
 * Read from .conf file.
 *****************************************************************************/
void
Conf::ReadConfTxt( const string& confTxtFilename )
{
    try
    {
        ModuleConfTxt confTxt;
        confTxt.Read( string(base::GetDataDir()) + string(confTxtFilename) );
    }
    catch (...)
    {
        CERROR << "ERROR: cannot read " << confTxtFilename << endl;
        EXIT( 1 );
    }
}

void
Conf::ModuleConfTxt::ProcessField( string& field, ifstream& conf )
{
    bool error = false;

    //---------------------  flat/sphere world  --------------------------------
    if ( field == "FlatObjectScale" and WORLD_CONF.mFlatWorld )
    {
        conf >> WORLD_CONF.mObjectScale;
        WORLD_CONF.mObjectScale3[XX] = \
        WORLD_CONF.mObjectScale3[YY] = \
        WORLD_CONF.mObjectScale3[ZZ] = WORLD_CONF.mObjectScale;
    }
    else if ( field == "SphereObjectScale" and not WORLD_CONF.mFlatWorld )
    {
        conf >> WORLD_CONF.mObjectScale;
        WORLD_CONF.mObjectScale3[XX] = \
        WORLD_CONF.mObjectScale3[YY] = \
        WORLD_CONF.mObjectScale3[ZZ] = WORLD_CONF.mObjectScale;
    }
    //--------------------  terrain model of flat world  -----------------------
    else if ( field == "FlatModelFile" )
    {
        conf >> WORLD_CONF.mFlatModelFile;
    }
    else if ( field == "FlatModelRotationList" )
    {
        WORLD_CONF.mFlatModelRotationList.Parse( conf );
    }
    else if ( field == "FlatModelScale" )
    {
        conf >> WORLD_CONF.mFlatModelScale;
    }
    else if ( field == "FlatModelOffset" )
    {
        conf >> WORLD_CONF.mFlatModelOffset[XX];
        conf >> WORLD_CONF.mFlatModelOffset[YY];
        conf >> WORLD_CONF.mFlatModelOffset[ZZ];
    }
    else if ( field == "FlatModelMeterSimRatio" )
    {
        conf >> WORLD_CONF.mFlatModelMeterSimRatio;
    }
    else if ( field == "FlatModelNW" )
    {
        fp f;
        conf >> f; WORLD_CONF.mFlatModelNW.mLat = Degree(f);  // SphereVertex
        conf >> f; WORLD_CONF.mFlatModelNW.mLon = Degree(f);
        conf >> f; WORLD_CONF.mFlatModelNW.mAlt = Meter(f);
    }
    else if ( field == "FlatModelSE" )
    {
        fp f;
        conf >> f; WORLD_CONF.mFlatModelSE.mLat = Degree(f);  // SphereVertex
        conf >> f; WORLD_CONF.mFlatModelSE.mLon = Degree(f);
        conf >> f; WORLD_CONF.mFlatModelSE.mAlt = Meter(f);
    }
    else if ( field == "FlatModelMeterOffset" )
    {
        fp f;
        conf >> f; WORLD_CONF.mFlatModelMeterOffset[XX] = Meter(f);
        conf >> f; WORLD_CONF.mFlatModelMeterOffset[YY] = Meter(f);
        conf >> f; WORLD_CONF.mFlatModelMeterOffset[ZZ] = Meter(f);
    }
    else if ( field == "FlatWorldMeterLowAlt" )
    {
        fp f; conf >> f; WORLD_CONF.mFlatModelMeterLowAlt = Meter(f);
    }
    else if ( field == "FlatModelMeterAltMin" )
    {
        fp f; conf >> f; WORLD_CONF.mFlatModelMeterAltMin = Meter(f);
    }
    else if ( field == "FlatModelMeterAltMax" )
    {
        fp f; conf >> f; WORLD_CONF.mFlatModelMeterAltMax = Meter(f);
    }

    // Errors?
    if ( error )
    {
        CERROR << "ERROR[conf]: parameter for " << field << " is invalid." << endl;
        EXIT( 1 );
    }
}

/*****************************************************************************
 * Print Conf configuration.
 *****************************************************************************/
ostream&
operator<<( ostream& strm, const Conf& conf )
{
    return strm;
}

} // namespace world
