/*****************************************************************************
 * $LastChangedDate: 2010-02-02 11:33:01 -0600 (Tue, 02 Feb 2010) $
 * @file
 * @author  Jim E. Brooks  http://www.palomino3d.org
 * @brief   Math test.
 *//*
 * LEGAL:   COPYRIGHT (C) 2006 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#define TESTING 1
#include "test.hh"
#include "base/module.hh"
using namespace base;
#include "math/module.hh"
#include "math/defs_trig.hh"
#include "math/defs_vector.hh"
#include "math/types_trig.hh"
#include "math/vertex.hh"
#include "math/funcs.hh"
#include "math/funcs_vector.hh"
using namespace math;

/*****************************************************************************
 * This asm code in C++ template returning args by reference looks suspicious,
 * so it's tested with different float types.
 *****************************************************************************/
template<typename FP>
void
SinCos_( Radian rad, FP& si, FP& co )
{
    asm("fsincos" : "=t" (co), "=u" (si) : "0" (rad));
}
bool
CheckSinCos( fp rad )
{
#if CPU_X86
    errno = 0;
 {
    fp si, co;
    SinCos_( rad, si, co );
    fp si2, co2;
    si2 = sin( rad );
    co2 = cos( rad );
    if ( not FP_EQ(si,si2) ) return false;
    if ( not FP_EQ(co,co2) ) return false;
 }

 {
    fpx si, co;
    SinCos_( rad, si, co );
    fpx si2, co2;
    si2 = sin( rad );
    co2 = cos( rad );
    if ( not FP_EQ(si,si2) ) return false;
    if ( not FP_EQ(co,co2) ) return false;
 }

 {
    long double si, co;
    SinCos_( rad, si, co );
    long double si2, co2;
    si2 = sin( rad );
    co2 = cos( rad );
    if ( not FP_EQ(si,si2) ) return false;
    if ( not FP_EQ(co,co2) ) return false;
 }

    return errno == 0;
#else
    return true;
#endif
}

/*****************************************************************************
 * Test SinCos().
 *****************************************************************************/
bool
TestSinCos( void )
{
    CERROR << "============================================\n";
    CERROR << "Testing SinCos() inline assembly. \n";
    CERROR << "============================================\n";

    if ( ! CheckSinCos( 0.0 ) ) goto Error;
    if ( ! CheckSinCos( 1.0 ) ) goto Error;
    if ( ! CheckSinCos( -1.0 ) ) goto Error;
    if ( ! CheckSinCos( -0.5 ) ) goto Error;
    if ( ! CheckSinCos( math::PI ) ) goto Error;
    if ( ! CheckSinCos( math::PI_DOUBLE ) ) goto Error;
    if ( ! CheckSinCos( -math::PI ) ) goto Error;
    if ( ! CheckSinCos( -math::PI_DOUBLE ) ) goto Error;
    if ( ! CheckSinCos( math::PI * 2.0 ) ) goto Error;
    if ( ! CheckSinCos( math::PI_DOUBLE * 2.0 ) ) goto Error;

    return true;

Error:
    return false;
}

/*****************************************************************************
 * Test FP_EQ().
 *
 * The faster varient FP_EQ(FP,FP) compares 3 fractional digits.
 *****************************************************************************/
bool
TestFP_EQ( void )
{
    CERROR << "============================================\n";
    CERROR << "Testing FP_EQ(). \n";
    CERROR << "============================================\n";

 {
    if ( not FP_EQ<double>( 100.0, 100.00001 ) ) return Fail();
    if ( FP_EQ<double>( 100.0, 100.1 ) ) return Fail();

    if ( not FP_EQ<fp>( 100.0, 100.001, 3 ) ) return Fail();
    if ( not FP_EQ<fp>( 100.0, 100.0001, 4 ) ) return Fail();
    if ( not FP_EQ<fp>( 100.0, 100.000001, 5 ) ) return Fail();
 }

 {
    float a = 1.2345;
    float b = 1.2345;
    if ( ! FP_EQ(a,b)   ) return Fail();
    if ( ! FP_EQ(a,b,4) ) return Fail();
    if ( ! FP_EQ(a,b,0) ) return Fail();
    if ( ! FP_EQ(a,b,1) ) return Fail();
    if ( ! FP_EQ(a,b,6) ) return Fail();
 }

 {
    double a = 1.2345;
    double b = 1.2345;
    if ( ! FP_EQ(a,b)   ) return Fail();
    if ( ! FP_EQ(a,b,4) ) return Fail();
    if ( ! FP_EQ(a,b,0) ) return Fail();
    if ( ! FP_EQ(a,b,1) ) return Fail();
    if ( ! FP_EQ(a,b,6) ) return Fail();
 }

 {
    float a = 1.22;
    float b = 1.23;
    if (   FP_EQ(a,b)   ) return Fail();  // compares 3 fractional digits
    if ( ! FP_EQ(a,b,0) ) return Fail();
    if ( ! FP_EQ(a,b,1) ) return Fail();
    if (   FP_EQ(a,b,3) ) return Fail();  // differs
 }

 {
    float a = 1.22;
    float b = 2.22;
    if ( FP_EQ(a,b)   ) return Fail();
    if ( FP_EQ(a,b,0) ) return Fail();
    if ( FP_EQ(a,b,1) ) return Fail();
    if ( FP_EQ(a,b,2) ) return Fail();
 }

 {
    float a = -3.21;
    float b =  3.21;
    if ( FP_EQ(a,b)   ) return Fail();
    if ( FP_EQ(a,b,0) ) return Fail();
    if ( FP_EQ(a,b,1) ) return Fail();
    if ( FP_EQ(a,b,2) ) return Fail();
 }

 {
    float a = 0.999999999;
    float b = 1.000000000;
    if ( ! FP_EQ(a,b)   ) return Fail();
    if ( ! FP_EQ(a,b,0) ) return Fail();
    if ( ! FP_EQ(a,b,1) ) return Fail();
    if ( ! FP_EQ(a,b,2) ) return Fail();
    if ( ! FP_EQ(a,b,5) ) return Fail();
 }

 {
    double a = 0.999999999;
    double b = 1.000000000;
    if ( ! FP_EQ(a,b)   ) return Fail();
    if ( ! FP_EQ(a,b,0) ) return Fail();
    if ( ! FP_EQ(a,b,1) ) return Fail();
    if ( ! FP_EQ(a,b,2) ) return Fail();
    if ( ! FP_EQ(a,b,5) ) return Fail();
 }

 {
    float a =  99999.9999;
    float b = 100000.0000;
    if ( ! FP_EQ(a,b)   ) return Fail();
    if ( ! FP_EQ(a,b,0) ) return Fail();
    if ( ! FP_EQ(a,b,1) ) return Fail();
    if ( ! FP_EQ(a,b,2) ) return Fail();
    if ( ! FP_EQ(a,b,5) ) return Fail();
 }

 {
    float a = -0.999999999;  // negative
    float b = 1.000000000;
    if ( FP_EQ(a,b)   ) return Fail();
    if ( FP_EQ(a,b,0) ) return Fail();
    if ( FP_EQ(a,b,1) ) return Fail();
    if ( FP_EQ(a,b,2) ) return Fail();
    if ( FP_EQ(a,b,5) ) return Fail();
 }

 {
    float a =   999.999;
    float b = -1000.000;  // negative
    if ( FP_EQ(a,b)   ) return Fail();
    if ( FP_EQ(a,b,0) ) return Fail();
    if ( FP_EQ(a,b,1) ) return Fail();
    if ( FP_EQ(a,b,2) ) return Fail();
    if ( FP_EQ(a,b,5) ) return Fail();
 }

    return true;
}

/*****************************************************************************
 * Test FP_LE().
 *****************************************************************************/
bool
TestFP_LE( void )
{
    CERROR << "============================================\n";
    CERROR << "Testing FP_LE(). \n";
    CERROR << "============================================\n";

 {
    if ( ! FP_LE<fp>( 10.0f, 20.0f ) ) return Fail();
    if ( ! FP_LE<fp>( 0.9f, 1.0f ) ) return Fail();
    if (   FP_LE<fp>( 1.1f, 1.0f ) ) return Fail();
    if ( ! FP_LE<fp>( 1.1f, 1.1f ) ) return Fail();
    if ( ! FP_LE<fp>( 1.09f, 1.1f ) ) return Fail();
    if ( ! FP_LE<fp>( -0.0000001f, 0.0f ) ) return Fail();  // close to equal
    if ( ! FP_LE<fp>( 1.000001f, 1.0f ) ) return Fail();  // close to equal
    if ( ! FP_LE<double>( 99.5, 99.500001 ) ) return Fail();  // close to equal
    if ( ! FP_LE<double>( 99.5, 99.499999 ) ) return Fail();  // close to equal
 }

    return true;
}

/*****************************************************************************
 * Test FP_GE().
 *****************************************************************************/
bool
TestFP_GE( void )
{
    CERROR << "============================================\n";
    CERROR << "Testing FP_GE(). \n";
    CERROR << "============================================\n";

 {
    if ( ! FP_GE<fp>( 20.0f, 10.0f ) ) return Fail();
    if ( ! FP_GE<fp>( 1.0f, 1.0f ) ) return Fail();
    if (   FP_GE<fp>( 1.0f, 2.0f ) ) return Fail();  // less
    if ( ! FP_GE<fp>( 5.0f, 1.0f ) ) return Fail();
    if ( ! FP_GE<fp>( -0.000001f, 0.0f ) ) return Fail();  // close to equal
    if ( ! FP_GE<fp>( 1.000001f, 1.0f ) ) return Fail();  // close to equal
    if ( ! FP_GE<double>( 99.5, 99.500001 ) ) return Fail();  // close to equal
    if ( ! FP_GE<double>( 99.5, 99.499999 ) ) return Fail();  // close to equal
 }

    return true;
}

/*****************************************************************************
 * Test IfInRange().
 *****************************************************************************/
bool
TestIfInRange( void )
{
    CERROR << "============================================\n";
    CERROR << "Testing IfInRange(). \n";
    CERROR << "============================================\n";

    if ( not IfInRange<int>( 10,20, 10,20 ) ) { return Fail(); }
    if ( not IfInRange<fp>( 10,20, 11,19 ) ) { return Fail(); }
    if ( not IfInRange<fp>( 10,20, 9,10 ) ) { return Fail(); }
    if ( not IfInRange<fp>( 10,20, -10,30 ) ) { return Fail(); }
    if ( not IfInRange<fp>( 10,20, 10,10 ) ) { return Fail(); }
    if ( not IfInRange<fp>( 10,20, 20,20 ) ) { return Fail(); }

    if ( IfInRange<fp>( 10,20, 0,1 ) ) { return Fail(); }
    if ( IfInRange<fp>( 10,20, 9,9 ) ) { return Fail(); }
    if ( IfInRange<fp>( 10,20, 21,99 ) ) { return Fail(); }
    if ( IfInRange<fp>( 10,20, -10,-20 ) ) { return Fail(); }

    return true;
}

/*****************************************************************************
 * Test IF_NAN().
 *****************************************************************************/
bool
TestIF_NAN( void )
{
    CERROR << "============================================\n";
    CERROR << "Testing IF_NAN(). \n";
    CERROR << "============================================\n";

    volatile fp f = 0.0f;
    volatile fp quotient = f / f;
    if ( not IF_NAN(quotient) ) { return Fail(); }
    if ( not IF_NAN( acos(999.0f) ) ) { return Fail(); }
    if ( IF_NAN(f) ) { return Fail(); }
    if ( IF_NAN(0.0f) ) { return Fail(); }

    return true;
}

/*****************************************************************************
 * Test Distance().
 *****************************************************************************/
bool
TestDistance( void )
{
    CERROR << "============================================\n";
    CERROR << "Testing Distance(). \n";
    CERROR << "============================================\n";

 {
    fp dist = Distance( Vector3( 1.0f, 0.0f, 0.0f ) );
    if ( not FP_EQ( dist, 1.0f ) ) { return Fail(); }
 }

 {
    fp dist = Distance( Vector3( 2.0f, 2.0f, 2.0f ) );
    const fp dist2 = SQRT( 4.0f + 4.0f + 4.0f );
    if ( not FP_EQ( dist, dist2 ) ) { return Fail(); }
 }

    return true;
}

/*****************************************************************************
 * Test Normalize().
 *****************************************************************************/
bool
TestNormalize( void )
{
    CERROR << "============================================\n";
    CERROR << "Testing Normalize(). \n";
    CERROR << "============================================\n";

 {
    const Vector3 v( 2.0f, 4.0f, -8.0f );
    const Vector3 n = Normalize( v );
    const fp      d = Distance( n );
    if ( not FP_EQ( d, 1.0f ) ) { return Fail(); }
 }

    return true;
}

/*****************************************************************************
 * Test CrossProduct() which makes a normal vector.
 *****************************************************************************/
bool
TestCrossProduct( void )
{
    CERROR << "============================================\n";
    CERROR << "Testing CrossProduct(). \n";
    CERROR << "============================================\n";

 {
    const Vector3 v0( 0.0f, 0.0f, 0.0f );  // origin
    const Vector3 v1( 1.0f, 0.0f, 0.0f );  // X
    const Vector3 v2( 0.0f, 1.0f, 0.0f );  // Y
    const Vector3 normal = CrossProduct( v1, v2, v0 );
    if ( (not FP_EQ( normal[XX], 0.0f ))
      or (not FP_EQ( normal[YY], 0.0f ))
      or (not FP_EQ( normal[ZZ], 1.0f )) )  // Z is normal to XY plane
        return Fail();
 }

 {
    const Vector3 v0( 0.0f, 0.0f, 0.0f );  // origin
    const Vector3 v1( 0.0f, 1.0f, 0.0f );  // Y
    const Vector3 v2( 0.0f, 0.0f, 1.0f );  // Z
    const Vector3 normal = CrossProduct( v1, v2, v0 );
    if ( (not FP_EQ( normal[XX], 1.0f ))
      or (not FP_EQ( normal[YY], 0.0f ))
      or (not FP_EQ( normal[ZZ], 0.0f )) )  // X is normal to YZ plane
        return Fail();
 }

    return true;
}

/*****************************************************************************
 * Test Angle3().
 *****************************************************************************/
bool
TestAngle3( void )
{
    CERROR << "============================================\n";
    CERROR << "Testing Angle3(). \n";
    CERROR << "============================================\n";

 {
    const Vector3 v0( 1.0f, 0.0f, 0.0f );
    const Vector3 v1( 0.0f, 1.0f, 0.0f );
    const Radian rad = Angle3( v0, v1 );
    if ( not FP_EQ( rad, Radian(RADIAN_90_DEG) ) ) { return Fail(); }
 }

 {
    const Vector3 v0( 1.0f, 0.0f, 0.0f );
    const Vector3 v1( -1.0f, 0.0f, 0.0f );
    const Radian rad = Angle3( v0, v1 );
    if ( not FP_EQ( rad, Radian(RADIAN_180_DEG) ) ) { return Fail(); }
 }

    return true;
}

/*****************************************************************************
 * Test Midpoint().
 *****************************************************************************/
bool
TestMidpoint( void )
{
    CERROR << "============================================\n";
    CERROR << "Testing Midpoint(). \n";
    CERROR << "============================================\n";

 {
    const Vector3 v0(  50.0f, 200.0f, -1000.0f );
    const Vector3 v1( 100.0f, 400.0f, -2000.0f );
    const Vector3 mid = Midpoint( v0, v1 );
    if ( not FP_EQ( mid[XX], 75.0f ) ) { return Fail(); }
    if ( not FP_EQ( mid[YY], 300.0f ) ) { return Fail(); }
    if ( not FP_EQ( mid[ZZ], -1500.0f ) ) { return Fail(); }
 }

    return true;
}

/*****************************************************************************
 * Test Midpoint().
 *****************************************************************************/
bool
TestInterpolate( void )
{
    CERROR << "============================================\n";
    CERROR << "Testing Interpolate(). \n";
    CERROR << "============================================\n";

 {
    const Vector3 v0(  50.0f, 200.0f, -1000.0f );
    const Vector3 v1( 100.0f, 400.0f, -2000.0f );
  //const Vector3 mid = Midpoint( v0, v1 );
    const Vector3 mid = Interpolate( v0, v1, 0.5f );  // equivalent
    if ( not FP_EQ( mid[XX], 75.0f ) ) { return Fail(); }
    if ( not FP_EQ( mid[YY], 300.0f ) ) { return Fail(); }
    if ( not FP_EQ( mid[ZZ], -1500.0f ) ) { return Fail(); }
 }

 {
    const Vector3 v0( 50.0f, 60.0f, 70.0f );
    const Vector3 v1( 100.0f, 100.0f, 100.0f );
    const Vector3 vi = Interpolate( v0, v1, 0.0f );
    if ( not FP_EQ( vi[XX], 50.0f ) ) { return Fail(); }
    if ( not FP_EQ( vi[YY], 60.0f ) ) { return Fail(); }
    if ( not FP_EQ( vi[ZZ], 70.0f ) ) { return Fail(); }
 }

 {
    const Vector3 v0( 50.0f, 60.0f, 70.0f );
    const Vector3 v1( 100.0f, 110.0f, 120.0f );
    const Vector3 vi = Interpolate( v0, v1, 1.0f );
    if ( not FP_EQ( vi[XX], 100.0f ) ) { return Fail(); }
    if ( not FP_EQ( vi[YY], 110.0f ) ) { return Fail(); }
    if ( not FP_EQ( vi[ZZ], 120.0f ) ) { return Fail(); }
 }

    return true;
}

/*****************************************************************************
 * main.
 *****************************************************************************/
int main( int argc, char** argv )
{
    RUN_TEST( "math", TEST_ERROR_MODE_ABORT,
              TestSinCos()
          and TestFP_EQ()
          and TestFP_LE()
          and TestFP_GE()
          and TestIfInRange()
          and TestIF_NAN()
          and TestDistance()
          and TestNormalize()
          and TestCrossProduct()
          and TestAngle3()
          and TestMidpoint()
          and TestInterpolate() )
}
