#!/usr/bin/env bash
# $LastChangedDate: 2010-02-05 08:31:09 -0600 (Fri, 05 Feb 2010) $
# Author: Jim Brooks  http://www.palomino3d.org
# Do some checks before compiling.
# cmake captures output to stderr and only prints messages
# from this script if it returns a non-zero exit code.
#-------------------------------------------------------------------------------

#-------------------------------------------------------------------------------
# Configuration:
LUA_DIRS="scripts tools"
#-------------------------------------------------------------------------------

TOP_SRC_DIR=$1
if [ "$TOP_SRC_DIR" = "" ]; then
    TOP_SRC_DIR=src
fi
[ -d $TOP_SRC_DIR ] || exit 1

# lwlint available?
if ( type -pf lwlint >/dev/null 2>&1 ); then
    echo "checks.sh: will use lwlint."
    HAVE_LWLINT=1
fi

GREP="grep -H -n"

Checks()
{
    SRC_DIR=$1

    # Run lwlint if available.
    if [ $HAVE_LWLINT ]; then
		if ( ls ${SRC_DIR}/*.cc >/dev/null 2>&1 ); then
            if ( ! lwlint ${SRC_DIR}/*.cc ${SRC_DIR}/*.hh ); then
                echo "ERROR: lwlint found errors." >/dev/stderr
                exit 1
            fi
		fi
    fi

    # "namespace global" singular is correct (minor).
    if ( $GREP -i "namespace globals" $SRC_DIR/*.* ); then
        echo "ERROR: namespace globals should be singular (minor)" >/dev/stderr
        exit 1
    fi

    # Catch: iter < end()
    if ( $GREP -i "iter.* < .*end[(]" $SRC_DIR/*.* ); then
        echo "ERROR: iter < end() is wrong!" >/dev/stderr
        exit 1
    fi

    # A habitual mistake that causes a buffer overrun is passing sizeof()
    # to InitArray() which expects an element count, not a byte count.
    if ( $GREP "InitArray.*sizeof" $SRC_DIR/*.* ); then  # not grep -q
        echo "ERROR: sizeof() cannot be passed to InitArray()" >/dev/stderr
        exit 1
    fi

    # Catch writing #if DEBUG2.  #if DEBUG == 2 is correct.
    if ( $GREP "if.*DEBUG2" $SRC_DIR/*.* ); then
        echo "ERROR: should be DEBUG==2" >/dev/stderr
        exit 1
    fi

    # Catch this mistake THREAD_CODE( Atomic::Add() )
    # because when compiled single-thread it expands nothing.
    if ( $GREP "\(^\|[^/]\)THREAD_CODE.*Atomic" $SRC_DIR/*.* ); then
        echo "ERROR: THREAD_CODE( Atomic ) expands nothing if ! COMPILE_THREADS" >/dev/stderr
        exit 1
    fi

    # Should be COMPILE_THREADS (plural).
    if ( $GREP "if COMPILE_THREAD$" $SRC_DIR/*.* ) || ( $GREP "if COMPILE_THREAD " $SRC_DIR/*.* ); then
        echo "ERROR: Should be COMPILE_THREADS (plural)." >/dev/stderr
        exit 1
    fi

    # shptr shouldn't be a reference.
    FILES=
    for FILE in $SRC_DIR/*.*
    do
        if ( echo $FILE | $GREP -q -v shptr ); then
            FILES="$FILES $FILE"
        fi

        # Allow: Class<shptr<Class> >&
        if ( echo $FILE | $GREP -q -v shptr ) && ( $GREP "[^<]shptr<.*>&" $FILES ); then
            echo "ERROR: shptr shouldn't be a reference" >/dev/stderr
            exit 1
        fi

        if ( echo $FILE | $GREP -q -v shptr ) && ( $GREP "[^<]RefPtr<.*>&" $FILES ); then
            echo "ERROR: RefPtr shouldn't be a reference" >/dev/stderr
            exit 1
        fi

    done

    # Catch using C headers instead of compatible superset C++ headers.
    # Can causes mysterious compile errors eg std::abs() will be missing variants.
    if ( $GREP "include..stdlib.h" $SRC_DIR/*.* ) \
    || ( $GREP "include..stdio.h" $SRC_DIR/*.* ) \
    || ( $GREP "include..math.h" $SRC_DIR/*.* ) \
    || ( $GREP "include..string.h" $SRC_DIR/*.* ) \
    || ( $GREP "include..errno.h" $SRC_DIR/*.* ) \
    then
        echo "ERROR: should use C++ headers instead of C headers" >/dev/stderr
        exit 1
    fi

    # Should write Count instead.
    if ( $GREP "Cnt" $SRC_DIR/*.* ); then
        echo "WARNING: Write Count instead." >/dev/stderr
        #exit 1
    fi
}

# For each subdir of SRC_DIR.
for SUBDIR in $TOP_SRC_DIR/*;
do
    if [ -d $SUBDIR ]; then
        if ( echo $SUBDIR | $GREP -q "deprecated\|lua_lang" ); then
            :
        else
            Checks $SUBDIR
        fi
    fi
done

# Check Lua scripts.
for LUA_DIR in $LUA_DIRS
do
    if [ ! -d $LUA_DIR ]; then
        echo "ERROR: LUA_DIR ${LUA_DIR} doesn't exist" >/dev/stderr
        exit 1
    fi

    # Find Lua scripts in this directory.
    LUA_SCRIPTS="`find ${LUA_DIR} -maxdepth 1 \( -name '*.lua' -or -name '*.py.lua' \)`"

    # Check Lua scripts using lint.lua.
    if [ "$LUA_SCRIPTS" != "" ]; then  # without args lint.lua would print help
        #echo "Checking for lint in ${LUA_SCRIPTS}" >/dev/stderr
        lua tools/lint.lua $LUA_SCRIPTS >/dev/stderr
        RC=$?
        if [ "$RC" = "1" ]; then
            exit 1
        elif [ "$RC" = "127" ]; then
            echo "Lua isn't installed, skipping lint checks (ok)." >/dev/stderr
            break  # no point in continuing
        fi
    fi
done

echo "Checks passed (ok)."
exit 0
